from samplerate import AudioSampleRate, audio_sample_rate
from soundfile import SoundFile
from io import BytesIO
from zounds.core import IdentityDimension, ArrayWithUnits
from timeseries import TimeDimension
from duration import Picoseconds
from samplerate import SampleRate


class AudioSamples(ArrayWithUnits):
    def __new__(cls, array, samplerate):
        if array.ndim == 1:
            dimensions = [TimeDimension(*samplerate)]
        elif array.ndim == 2:
            dimensions = [TimeDimension(*samplerate), IdentityDimension()]
        else:
            raise ValueError(
                    'array must be one (mono) or two (multi-channel) dimensions')

        if not isinstance(samplerate, AudioSampleRate):
            raise TypeError('samplerate should be an AudioSampleRate instance')

        return ArrayWithUnits.__new__(cls, array, dimensions)

    def __add__(self, other):
        try:
            if self.samplerate != other.samplerate:
                raise ValueError(
                        'Samplerates must match, but they were '
                        '{self.samplerate} and {other.samplerate}'
                        .format(**locals()))
        except AttributeError:
            pass
        return super(AudioSamples, self).__add__(other)

    def kwargs(self):
        return {'samplerate': self.samplerate}

    def sum(self, axis=None, dtype=None, **kwargs):
        result = super(AudioSamples, self).sum(axis, dtype, **kwargs)
        if self.ndim == 2 and axis == 1:
            return AudioSamples(result, self.samplerate)
        else:
            return result

    @property
    def samples_per_second(self):
        return int(Picoseconds(int(1e12)) / self.frequency)

    @property
    def duration_in_seconds(self):
        return self.duration / Picoseconds(int(1e12))

    @property
    def samplerate(self):
        return SampleRate(self.frequency, self.duration)

    @property
    def overlap(self):
        return self.samplerate.overlap

    @property
    def span(self):
        return self.dimensions[0].span

    @property
    def end(self):
        return self.dimensions[0].end

    @property
    def frequency(self):
        return self.dimensions[0].frequency

    @property
    def duration(self):
        return self.dimensions[0].duration

    @classmethod
    def from_example(cls, arr, example):
        return cls(arr, example.samplerate)

    @property
    def channels(self):
        if len(self.shape) == 1:
            return 1
        return self.shape[1]

    @property
    def samplerate(self):
        return audio_sample_rate(self.samples_per_second)

    @property
    def mono(self):
        if self.channels == 1:
            return self
        x = self.sum(axis=1) * 0.5
        y = x * 0.5
        return AudioSamples(y, self.samplerate)

    def encode(self, flo=None, fmt='WAV', subtype='PCM_16'):
        flo = flo or BytesIO()
        with SoundFile(
                flo,
                mode='w',
                channels=self.channels,
                format=fmt,
                subtype=subtype,
                samplerate=self.samples_per_second) as f:
            f.write(self)
        flo.seek(0)
        return flo



from __future__ import division

import numpy as np
from numpy.lib.stride_tricks import as_strided as ast

import pyximport
pyximport.install()
from countbits import *


def norm_shape(shape):
    """
    Normalize numpy array shapes so they're always expressed as a tuple,
    even for one-dimensional shapes.

    Parameters
        shape - an int, or a tuple of ints

    Returns
        a shape tuple
    """
    try:
        i = int(shape)
        return (i,)
    except TypeError:
        # shape was not a number
        pass

    try:
        t = tuple(shape)
        return t
    except TypeError:
        # shape was not iterable
        pass

    raise TypeError('shape must be an int, or a tuple of ints')


def flatten2d(arr):
    ls = len(arr.shape)
    if 1 == ls:
        raise ValueError('Cannot turn 1d array into 2d array')
    elif 2 == ls:
        return arr
    else:
        return arr.reshape((arr.shape[0], np.product(arr.shape[1:])))


def downsampled_shape(shape, factor):
    """
    Return the new shape of an array with shape, once downsampled
    by factor.
    """
    return tuple((np.array(shape) / factor).astype(np.int32))


def downsample(arr, factor, method=np.max):
    """
    Downsample an n-dimensional array

    Parameters
        arr    - the array to be downsampled
        factor - the downsampling factor. If factor is an integer, it is assumed
                 that the array will be downsampled by a constant factor in every
                 dimension. If factor is a tuple, the array will be downsampled
                 by different factors in each dimension. If factor is a tuple
                 whose size is less than the number of dimensions in arr, it
                 is assumed that the downsampling will only be applied to the
                 last len(factor) dimensions of arr.
        method - the method used to combine chunks into single values, e.g.
                 mean, or max.
    """
    if isinstance(factor, int):
        # factor is an integer, so we'll downsample by a constant factor over
        # all dimensions of the input array
        factor = (factor,) * arr.ndim
        lf = arr.ndim
    else:
        # factor is a tuple, so we'll downsample over the last n dimensions of
        # arr by the factors specified
        factor = norm_shape(factor)
        lf = len(factor)
        if lf > arr.ndim:
            raise ValueError( \
                    'The number of factors must be less than or equal to the number of dimensions in arr')

    # the axes over which to apply the reduction
    axes = -np.arange(1, lf + 1)
    # the window size in each dimension of arr
    window_size = ((1,) * (arr.ndim - lf)) + factor
    # get non-overlapping windows whose dimensions are specified by factor
    windowed = sliding_window(arr, window_size, flatten=False)
    # apply the reduction and remove any extraneous dimensions
    return np.apply_over_axes(method, windowed, axes).squeeze()


def safe_log(a):
    """
    Return the element-wise log of an array, checking for negative
    array elements and avoiding divide-by-zero errors.
    """
    if np.any([a < 0]):
        raise ValueError('array contains negative components')

    return np.log(a + 1e-12)


def euclidean_norm(a):
    """
    Return the euclidean norm of each data example
    """
    if 1 == len(a.shape):
        return np.linalg.norm(a)

    return np.sum(np.abs(a) ** 2, axis=-1) ** (1. / 2)


def safe_unit_norm(a):
    """
    Ensure that the vector or vectors have unit norm
    """
    if 1 == len(a.shape):
        n = np.linalg.norm(a)
        if n:
            return a / n
        return a

    norm = np.sum(np.abs(a) ** 2, axis=-1) ** (1. / 2)

    # Dividing by a norm of zero will cause a warning to be issued. Set those
    # values to another number. It doesn't matter what, since we'll be dividing
    # a vector of zeros by the number, and 0 / N always equals 0.
    norm[norm == 0] = -1e12
    return a / norm[:, np.newaxis]


def binarize(arr, thresh=0.5):
    arr[arr < thresh] = 0
    arr[arr >= thresh] = 1
    return arr


def pad(a, desiredlength):
    """
    Pad an n-dimensional numpy array with zeros along the zero-th dimension
    so that it is the desired length.  Return it unchanged if it is greater
    than or equal to the desired length
    """

    if len(a) >= desiredlength:
        return a

    islist = isinstance(a, list)
    a = np.array(a)
    diff = desiredlength - len(a)
    shape = list(a.shape)
    shape[0] = diff

    padded = np.concatenate([a, np.zeros(shape, dtype=a.dtype)])
    return padded.tolist() if islist else padded


def _wpad(l, windowsize, stepsize):
    """
    Parameters
        l - The length of the input array
        windowsize - the size of each window of samples
        stepsize - the number of samples to move the window each step

    Returns
        The length the input array should be so that no samples are leftover
    """
    if l <= windowsize:
        return windowsize

    nsteps = ((l // stepsize) * stepsize)
    overlap = (windowsize - stepsize)
    if overlap:
        return nsteps + overlap

    diff = (l - nsteps)
    left = max(0, windowsize - diff)
    return l + left if diff else l


def _wcut(l, windowsize, stepsize):
    """
    Parameters
        l - The length of the input array
        windowsize - the size of each window of samples
        stepsize - the number of samples to move the window each step

    Returns
        The length the input array should be so that leftover samples are ignored
    """
    end = l - windowsize
    if l <= windowsize:
        return 0, 0
    elif end % stepsize:
        l = windowsize + ((end // stepsize) * stepsize)

    return l, l - (windowsize - stepsize)


def windowed(a, windowsize, stepsize=None, dopad=False):
    """
    Parameters
        a          - the input array to restructure into overlapping windows
        windowsize - the size of each window of samples
        stepsize   - the number of samples to shift the window each step. If not
                     specified, this defaults to windowsize
        dopad      - If false (default), leftover samples are returned seperately.
                     If true, the input array is padded with zeros so that all
                     samples are used.
    """
    if windowsize < 1:
        raise ValueError('windowsize must be greater than or equal to one')

    if stepsize is None:
        stepsize = windowsize

    if stepsize < 1:
        raise ValueError('stepsize must be greater than or equal to one')

    if not a.flags['C_CONTIGUOUS']:
        a = a.copy()

    if windowsize == 1 and stepsize == 1:
        # A windowsize and stepsize of one mean that no windowing is necessary.
        # Return the array unchanged.
        return np.zeros((0,) + a.shape[1:], dtype=a.dtype), a

    if windowsize == 1 and stepsize > 1:
        return np.zeros(0, dtype=a.dtype), a[::stepsize]

    # the original length of the input array
    l = a.shape[0]

    if dopad:
        p = _wpad(l, windowsize, stepsize)
        # pad the array with enough zeros so that there are no leftover samples
        a = pad(a, p)
        # no leftovers; an empty array
        leftover = np.zeros((0,) + a.shape[1:], dtype=a.dtype)
    else:
        # cut the array so that any leftover samples are returned seperately
        c, lc = _wcut(l, windowsize, stepsize)
        leftover = a[lc:]
        a = a[:c]

    if 0 == a.shape[0]:
        return leftover, np.zeros(a.shape, dtype=a.dtype)

    n = 1 + (a.shape[0] - windowsize) // (stepsize)
    s = a.strides[0]
    newshape = (n, windowsize) + a.shape[1:]
    newstrides = (stepsize * s, s) + a.strides[1:]

    out = np.ndarray.__new__( \
            np.ndarray,
            strides=newstrides,
            shape=newshape,
            buffer=a,
            dtype=a.dtype)

    return leftover, out


def sliding_window(a, ws, ss=None, flatten=True):
    """
    Return a sliding window over a in any number of dimensions

    Parameters:
        a  - an n-dimensional numpy array
        ws - an int (a is 1D) or tuple (a is 2D or greater) representing the size
             of each dimension of the window
        ss - an int (a is 1D) or tuple (a is 2D or greater) representing the
             amount to slide the window in each dimension. If not specified, it
             defaults to ws.
        flatten - if True, all slices are flattened, otherwise, there is an
                  extra dimension for each dimension of the input.

    Returns
        an array containing each n-dimensional window from a
    """

    if None is ss:
        # ss was not provided. the windows will not overlap in any direction.
        ss = ws
    ws = norm_shape(ws)
    ss = norm_shape(ss)

    # convert ws, ss, and a.shape to numpy arrays so that we can do math in every 
    # dimension at once.
    ws = np.array(ws)
    ss = np.array(ss)
    shape = np.array(a.shape)

    # ensure that ws, ss, and a.shape all have the same number of dimensions
    ls = [len(shape), len(ws), len(ss)]
    if 1 != len(set(ls)):
        raise ValueError( \
                'a.shape, ws and ss must all have the same length. They were %s' % str(
                    ls))

    # ensure that ws is smaller than a in every dimension
    if np.any(ws > shape):
        raise ValueError( \
                'ws cannot be larger than a in any dimension.\
 a.shape was %s and ws was %s' % (str(a.shape), str(ws)))

    # how many slices will there be in each dimension?
    newshape = norm_shape(((shape - ws) // ss) + 1)
    # the shape of the strided array will be the number of slices in each dimension
    # plus the shape of the window (tuple addition)
    newshape += norm_shape(ws)
    # the strides tuple will be the array's strides multiplied by step size, plus
    # the array's strides (tuple addition)
    newstrides = norm_shape(np.array(a.strides) * ss) + a.strides


    strided = ast(a, shape=newshape, strides=newstrides)
    if not flatten:
        return strided

    # Collapse strided so that it has one more dimension than the window.  I.e.,
    # the new array is a flat list of slices.
    meat = len(ws) if ws.shape else 0
    firstdim = (np.product(newshape[:-meat]),) if ws.shape else ()
    dim = firstdim + (newshape[-meat:])
    # remove any dimensions with size 1
    dim = filter(lambda i: i != 1, dim)
    return strided.reshape(dim)


class TypeCodes(object):
    _bits = [8, 16, 32, 64]
    _np_types = [np.uint8, np.uint16, np.uint32, np.uint64]
    _type_codes = ['B', 'H', 'L', 'Q']

    @classmethod
    def _fromto(cls, f, t, v):
        return t[f.index(v)]

    @classmethod
    def _whichlist(cls, v):

        if isinstance(v, int):
            return cls._bits

        if isinstance(v, str):
            return cls._type_codes

        if isinstance(v, type) or isinstance(v, np.dtype):
            return cls._np_types

        raise ValueError('%s is not a valid key' % v)

    @classmethod
    def bits(cls, v):
        """
        Get the number of bits from a numpy dtype or typecode
        """
        return cls._fromto(cls._whichlist(v), cls._bits, v)

    @classmethod
    def bytes(cls, v):
        return cls.bits(v) // 8

    @classmethod
    def np_dtype(cls, v):
        """
        Get a numpy dtype from a typecode or number of bits
        """
        return cls._fromto(cls._whichlist(v), cls._np_types, v)

    @classmethod
    def type_code(cls, v):
        """
        Get a typecode from a numpy dtype or a number of bits
        """
        return cls._fromto(cls._whichlist(v), cls._type_codes, v)


class Growable(object):
    """
    A thin wrapper around a numpy array that allows it to be treated like a
    dynamically-sized object.  For numeric types, this should be more memory
    efficient than using a list, since members are represented by contiguous
    blocks of memory instead of PyObjects.
    """

    def __init__(self, data, position=0, growth_rate=1):
        """__init__

        :param data: A numpy array to treat as dynamically resizable. Can have \
        length 0

        :param position: The logical size of the array.  This can be smaller, \
        equal to, or larger than the array passed in as data

        :param growth_rate: The amount by which the array will grow when \
        necessary.  A growth rate of 1 means that the array will double in \
        size each time.
        """
        if growth_rate <= 0:
            raise ValueError('growth_rate must be greater than zero')

        object.__init__(self)
        self._data = data
        self._position = position
        self._growth_rate = growth_rate

    @property
    def data(self):
        """
        Return the numpy array wrapped by this instance.  Note that the array
        may be larger or smaller than logical_size
        """
        return self._data

    @property
    def logical_size(self):
        """
        Return the size of the initialized part of the array.  Values outside
        this boundary are invalid.
        """
        return self._position

    @property
    def physical_size(self):
        """
        The actual size of the wrapped array
        """
        return self._data.shape[0]

    @property
    def logical_data(self):
        """
        Return only the initialized part of the array
        """
        return self._data[:self.logical_size]

    def _tmp_size(self):
        n_items = int(self._data.shape[0] * self._growth_rate)
        # require that n_items is at least one
        return 1 if not n_items else n_items

    def _tmp(self, amt=None):
        n_items = self._tmp_size() if None is amt else amt
        return np.ndarray( \
                (n_items,) + self._data.shape[1:], dtype=self._data.dtype)

    def _grow(self, amt=None):
        new_mem = self._tmp(amt=amt)
        self._data = np.concatenate([self._data, new_mem])

    def append(self, item):
        """
        append a single item to the array, growing the wrapped numpy array
        if necessary
        """
        try:
            self._data[self._position] = item
        except IndexError:
            self._grow()
            self._data[self._position] = item
        self._position += 1
        return self

    def extend(self, items):
        """
        extend the numpy array with multiple items, growing the wrapped array
        if necessary
        """
        items = np.array(items)
        pos = items.shape[0] + self.logical_size
        if pos > self.physical_size:
            # more physical memory is needed than has been allocated so far
            amt = self._tmp_size()
            if self.physical_size + amt < pos:
                # growing the memory by the prescribed amount still won't 
                # make enough room. Allocate exactly as much memory as is needed
                # to accommodate items
                amt = pos - self.physical_size

            # actually grow the array
            self._grow(amt=amt)

        stop = self._position + items.shape[0]
        self._data[self._position: stop] = items
        self._position += items.shape[0]
        return self


def hamming_distance(a, b):
    """
    a - scalar
    b - array of scalars
    """
    return count_bits(a ^ b)


def packed_hamming_distance(a, b):
    """
    Interpret a as a "packed" scalar, i.e. an n-bit number where n may not be
    a power of 2. E.g., a 250-bit number would be represented by 4 64-bit integers.

    Interpret b as an array of "packed" scalars. Its second dimension should be
    the same length as a.
    """
    xored = a ^ b
    bitcount = count_bits(xored.ravel())
    return bitcount.reshape(xored.shape).sum(1)


def jaccard_distance(a, b):
    """
    a - scalar
    b - array of scalars
    """
    intersection = count_bits(a & b)
    union = count_bits(a | b)
    return 1 - (intersection / union)


# TODO: Write tests
class BKTree(object):
    """
    Take advantage of the triangle inequality.  Build a depth-one tree
    where each branch contains the indices of all hashes which are distance n
    from a seed value.

    When querying, take the distance between the query and seed values.  Begin
    with the branch for that distance, and proceed outward.
    """

    def __init__(self, hashes=None, distance=hamming_distance):
        object.__init__(self)
        self.hashes = hashes
        self.distance = distance
        self.seed = self.hashes[np.random.randint(0, len(hashes))]
        self._build()

    def _build(self):
        d = self.distance(self.seed, self.hashes)
        self.tree = dict((dist, np.where(d == dist)[0]) for dist in set(d))

    def append(self, hashes):
        d = self.distance(self.seed, self.hashes)
        tree = dict((dist, np.where(d == dist)[0]) for dist in set(d))
        for k, v in tree.iteritems():
            try:
                self.tree[k] = np.concatenate([v, self.tree[k]])
            except KeyError:
                self.tree[k] = v

    def query(self, query, nresults):
        d = self.distance(query, [self.seed])[0]
        results = np.zeros(nresults, dtype=np.uint32)
        resultspos = 0
        keys = np.array(self.tree.keys())
        ordered = list(keys[np.argsort(np.abs(keys - d))])
        while ordered and resultspos < nresults:
            dist = ordered.pop(0)
            r = self.tree[dist]
            lr = len(r)
            stop = min(nresults, resultspos + lr)
            results[resultspos: stop] = r[:stop - resultspos]
            resultspos += (stop - resultspos)

        return results



import numpy as np
from zounds.nputil import sliding_window, windowed
from dimensions import IdentityDimension
import copy


class CustomSlice(object):
    def __init__(self):
        super(CustomSlice, self).__init__()


class ArrayWithUnits(np.ndarray):
    def __new__(cls, arr, dimensions):

        if arr.ndim != len(dimensions):
            raise ValueError(
                'arr.ndim and len(dimensions) must match.  '
                'They were {arr.shape} and {dimensions}'.format(**locals()))

        def dim_map(d):
            return IdentityDimension() if d is None else copy.deepcopy(d)

        obj = np.asarray(arr).view(cls)
        obj.dimensions = tuple(map(dim_map, dimensions))

        for dim, size in zip(obj.dimensions, obj.shape):
            try:
                dim.size = size
            except AttributeError:
                pass
            try:
                dim.validate(size)
            except AttributeError:
                pass

        return obj

    def kwargs(self):
        return self.__dict__

    def concatenate(self, other):
        if self.dimensions == other.dimensions:
            return self.from_example(np.concatenate([self, other]), self)
        else:
            raise ValueError('All dimensions must match to concatenate')

    def reshape(self, shape, order='C'):
        raw = np.asarray(self)
        return np.reshape(raw, shape, order)

    @classmethod
    def concat(cls, arrs, axis=0):
        for arr in arrs[1:]:
            if arr.dimensions != arrs[0].dimensions:
                raise ValueError('All dimensions must match')
        return cls.from_example(np.concatenate(arrs, axis=axis), arrs[0])

    @classmethod
    def from_example(cls, data, example):
        return ArrayWithUnits(data, example.dimensions)

    def sum(self, axis=None, dtype=None, **kwargs):
        result = super(ArrayWithUnits, self).sum(axis, dtype, **kwargs)
        if axis is not None:
            new_dims = list(self.dimensions)
            new_dims.pop(axis)
            return ArrayWithUnits(result, new_dims)
        else:
            # we have a scalar
            return result

    def max(self, axis=None, out=None):
        result = super(ArrayWithUnits, self).max(axis=axis, out=out)
        if axis is not None:
            new_dims = list(self.dimensions)
            new_dims.pop(axis)
            return ArrayWithUnits(result, new_dims)
        else:
            # we have a scalar
            return result

    def dot(self, b):
        result = super(ArrayWithUnits, self).dot(b)
        return self.__class__(result, self.dimensions[:result.ndim])

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.dimensions = getattr(obj, 'dimensions', None)

    def __array_wrap__(self, obj, context=None):
        if len(self.dimensions) != obj.ndim:
            if obj.ndim == 0:
                return obj[0]
            return np.asarray(obj)
        return np.ndarray.__array_wrap__(self, obj, context)

    def sliding_window(self, windowsize, stepsize=None):
        ws = tuple(self._compute_span(windowsize))
        ss = tuple(self._compute_span(stepsize)) if stepsize else ws
        result = sliding_window(self, ws, ss)

        try:
            new_dims = tuple(self._compute_new_dims(result, ws, ss))
        except ValueError:
            new_dims = [IdentityDimension()] * result.ndim

        return ArrayWithUnits(result, new_dims)

    def _sliding_window_integer_slices(self, windowsize, stepsize=None):
        windowsize = [windowsize] + [slice(None) for _ in self.dimensions[1:]]
        ws = tuple(self._compute_span(windowsize))
        if stepsize:
            stepsize = [stepsize] + [slice(None) for _ in self.dimensions[1:]]
            ss = tuple(self._compute_span(stepsize))
        else:
            ss = ws

        print ws, ss
        return ws, ss

    def sliding_window_with_leftovers(
            self, windowsize, stepsize=None, dopad=False):

        ws, ss = \
            self._sliding_window_integer_slices(windowsize, stepsize)

        leftovers, result = windowed(self, ws[0], ss[0], dopad)

        if not result.size:
            return self, ArrayWithUnits(result, self.dimensions)

        try:
            new_dims = tuple(self._compute_new_dims(result, ws, ss))
        except ValueError:
            new_dims = [IdentityDimension()] * result.ndim

        return leftovers, ArrayWithUnits(result, new_dims)

    def _compute_new_dims(self, windowed, ws, ss):
        for dimension, size, w, s in zip(self.dimensions, self.shape, ws, ss):
            try:
                modified = dimension.modified_dimension(size, w, s)
                for m in modified:
                    yield m
            except NotImplementedError:
                yield dimension

    def _compute_span(self, index):
        for sl in self._compute_indices(index):
            try:
                yield sl.stop - sl.start
            except (AttributeError, TypeError):
                yield sl

    def _compute_indices(self, index):
        dims_pos = 0
        for sl in index:
            if sl is None \
                    or isinstance(sl, int) \
                    or isinstance(sl, slice) \
                    or isinstance(sl, list):
                # burn one
                dims_pos += 1
                yield sl
            elif sl is Ellipsis:
                dims_pos += len(self.dimensions) - (len(index) - 1)
                yield Ellipsis
            else:
                dim = self.dimensions[dims_pos]
                yield dim.integer_based_slice(sl)
                dims_pos += 1

    def _new_dims(self, index, new_arr):
        dims_pos = 0
        shape_pos = 0
        for sl in index:
            if sl is None:
                # additional dimension via np.newaxis
                yield IdentityDimension()
            elif isinstance(sl, int):
                # burn a dimension
                dims_pos += 1
                shape_pos += 1
            elif isinstance(sl, list):
                dims_pos += 1
                shape_pos += 1
                yield IdentityDimension()
            elif sl is Ellipsis:
                ellipsis_size = len(self.dimensions) - (len(index) - 1)
                for i in xrange(ellipsis_size):
                    yield self.dimensions[dims_pos]
                    dims_pos += 1
                    shape_pos += 1
            else:
                try:
                    dim = self.dimensions[dims_pos]
                    shape = new_arr.shape[shape_pos]
                    yield dim.metaslice(sl, shape)
                except IndexError:
                    yield dim
                dims_pos += 1
                shape_pos += 1

        # Return any leftover dimensions
        for dim in self.dimensions[dims_pos:]:
            yield dim

    def __gt__(self, other):
        return np.asarray(self).__gt__(other)

    def __ge__(self, other):
        return np.asarray(self).__ge__(other)

    def __lt__(self, other):
        return np.asarray(self).__lt__(other)

    def __le__(self, other):
        return np.asarray(self).__le__(other)

    def __eq__(self, other):
        return np.asarray(self).__eq__(other)

    def _tuplify(self, a):
        if isinstance(a, list):
            t = set(map(lambda x: x.__class__, a))
            if len(t) > 1:
                raise ValueError('a must be homogeneous')
            t = list(t)[0]
            if t == slice:
                return a
            else:
                return a,
        if isinstance(a, np.ndarray) and a.dtype == np.bool:
            return a,
        try:
            return tuple(a)
        except TypeError:
            return a,

    def __getslice__(self, i, j):
        return self.__getitem__(slice(i, j))

    def __setitem__(self, index, value):
        index = self._tuplify(index)
        indices = tuple(self._compute_indices(index))
        super(ArrayWithUnits, self).__setitem__(indices, value)

    def __getitem__(self, index):
        if isinstance(index, np.ndarray) \
                and index.dtype == np.bool:
            return np.asarray(self)[index]

        if self.ndim == 1 and isinstance(index, int):
            return np.asarray(self)[index]

        index = self._tuplify(index)
        indices = tuple(self._compute_indices(index))
        print indices
        arr = super(ArrayWithUnits, self).__getitem__(indices)
        new_dims = tuple(self._new_dims(index, arr))
        return ArrayWithUnits(arr, new_dims)



import numpy as np
import unittest2
from duration import \
    Picoseconds, Milliseconds, Seconds, Microseconds, Nanoseconds, Hours
from timeseries import TimeSlice, TimeDimension
from zounds.core import IdentityDimension, ArrayWithUnits


class ConvenienceClassTests(unittest2.TestCase):
    def test_seconds_equal(self):
        a = Seconds(1)
        b = np.timedelta64(1, 's')
        self.assertEqual(a, b)

    def test_milliseconds_equal(self):
        a = Milliseconds(1000)
        b = np.timedelta64(1000, 'ms')
        self.assertEqual(a, b)

    def test_microseconds_equal(self):
        a = Microseconds(10)
        b = np.timedelta64(10, 'us')
        self.assertEqual(a, b)

    def test_different_units_equal(self):
        self.assertEqual(Seconds(1), Milliseconds(1000))


class TimeSliceTests(unittest2.TestCase):
    def test_can_repr_empty_slice(self):
        r = repr(TimeSlice())
        self.assertIsNotNone(r)

    def test_can_repr_open_ended_slice(self):
        r = repr(TimeSlice(start=Seconds(2)))
        self.assertIsNotNone(r)

    def test_raises_if_duration_is_not_timedelta_instance(self):
        self.assertRaises(ValueError, lambda: TimeSlice(1))

    def test_raises_if_start_is_provided_and_is_not_timedelta_instance(self):
        self.assertRaises(
                ValueError, lambda: TimeSlice(Nanoseconds(100), 1))

    def test_can_instantiate_time_slice_instance_without_start_argument(self):
        duration = Seconds(100)
        ts = TimeSlice(duration)
        self.assertEqual(duration, ts.duration)
        self.assertEqual(Seconds(0), ts.start)

    def test_can_instantiate_time_slice_instance_with_start_argument(self):
        duration = Microseconds(1000)
        start = Hours(1)
        ts = TimeSlice(duration, start=start)
        self.assertEqual(duration, ts.duration)
        self.assertEqual(start, ts.start)

    def test_can_intersect_two_time_slices(self):
        ts1 = TimeSlice(Seconds(100), start=Seconds(100))
        ts2 = TimeSlice(Seconds(100), start=Seconds(101))
        intersection = ts1 & ts2
        self.assertEqual(Seconds(99), intersection.duration)

    def test_can_find_null_intersection(self):
        ts1 = TimeSlice(Seconds(100), start=Seconds(100))
        ts2 = TimeSlice(Seconds(100), start=Seconds(200))
        intersection = ts1 & ts2
        self.assertEqual(Seconds(0), intersection.duration)

    def test_does_not_contain_point_in_time_before(self):
        ts = TimeSlice(Seconds(100), start=Seconds(200))
        self.assertFalse(Seconds(10) in ts)

    def test_contains_point_in_time_during(self):
        ts = TimeSlice(Seconds(100), start=Seconds(200))
        self.assertTrue(Seconds(210) in ts)

    def test_does_not_contain_point_in_time_after(self):
        ts = TimeSlice(Seconds(100), start=Seconds(200))
        self.assertFalse(Seconds(310) in ts)

    def test_does_not_contain_slice_completely_before(self):
        ts1 = TimeSlice(Seconds(100), start=Seconds(200))
        ts2 = TimeSlice(Seconds(10), Seconds(12))
        self.assertFalse(ts2 in ts1)

    def test_does_not_contain_slice_beginning_before(self):
        ts1 = TimeSlice(Seconds(100), start=Seconds(200))
        ts2 = TimeSlice(Seconds(50), Seconds(190))
        self.assertFalse(ts2 in ts1)

    def test_contains_slice(self):
        ts1 = TimeSlice(Seconds(100), start=Seconds(200))
        ts2 = TimeSlice(Seconds(10), Seconds(250))
        self.assertTrue(ts2 in ts1)

    def test_does_not_contain_slice_completely_after(self):
        ts1 = TimeSlice(Seconds(100), start=Seconds(200))
        ts2 = TimeSlice(Seconds(100), Seconds(310))
        self.assertFalse(ts2 in ts1)

    def test_does_not_contain_slice_beginning_after(self):
        ts1 = TimeSlice(Seconds(100), start=Seconds(200))
        ts2 = TimeSlice(Seconds(100), Seconds(210))
        self.assertFalse(ts2 in ts1)

    def test_raises_value_error_if_item_is_not_timedelta_or_timeslice(self):
        ts1 = TimeSlice(Seconds(100), start=Seconds(200))
        self.assertRaises(ValueError, lambda: 's' in ts1)

    def test_eq_when_start_and_duration_equal(self):
        ts1 = TimeSlice(Seconds(2), start=Seconds(2))
        ts2 = TimeSlice(Seconds(2), start=Seconds(2))
        self.assertEqual(ts1, ts2)

    def test_ne_when_durations_differ(self):
        ts1 = TimeSlice(Seconds(2), start=Seconds(2))
        ts2 = TimeSlice(Seconds(3), start=Seconds(2))
        self.assertNotEqual(ts1, ts2)

    def test_ne_when_starts_differ(self):
        ts1 = TimeSlice(Seconds(2), start=Seconds(2))
        ts2 = TimeSlice(Seconds(2), start=Seconds(3))
        self.assertNotEqual(ts1, ts2)


class TimeSeriesTests(unittest2.TestCase):

    def test_can_slice_time_series_with_time_slice(self):
        arr = np.arange(10)
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        sl = TimeSlice(Seconds(2), start=Seconds(2))
        ts2 = ts[sl]
        self.assertEqual(2, len(ts2))

    def test_can_slice_time_series_with_open_ended_time_slice(self):
        arr = np.arange(10)
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        sl = TimeSlice(None, start=Seconds(2))
        ts2 = ts[sl]
        self.assertEqual(8, len(ts2))

    def test_can_index_constant_rate_time_series_with_integer_index(self):
        arr = np.arange(10)
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        ts2 = ts[5]
        self.assertEqual(5, ts2)

    def test_can_index_2d_time_series_with_single_integer_index(self):
        arr = np.random.random_sample((10, 3))
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq), IdentityDimension()])
        ts2 = ts[5]
        self.assertEqual((3,), ts2.shape)

    def test_can_mix_time_slice_and_integer_indices(self):
        arr = np.ones((10, 5))
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq), IdentityDimension()])
        sl = TimeSlice(duration=Seconds(5), start=Seconds(5))
        ts2 = ts[sl, 2:]
        self.assertEqual((5, 3), ts2.shape)

    def test_can_slice_constant_rate_time_series_with_integer_indices(self):
        arr = np.arange(10)
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        ts2 = ts[:5]
        self.assertEqual(5, len(ts2))
        self.assertIsInstance(ts2, ArrayWithUnits)

    def test_can_add_constant_factor_to_time_series(self):
        arr = np.arange(10)
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        ts2 = ts + 10
        self.assertTrue(np.all(np.arange(10, 20) == ts2))
        self.assertIsInstance(ts2, ArrayWithUnits)

    def test_get_index_error_when_using_out_of_range_int_index(self):
        arr = np.arange(10)
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        self.assertRaises(IndexError, lambda: ts[100])

    def test_get_empty_time_series_when_using_out_of_range_time_slice(self):
        arr = np.arange(10)
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        sl = TimeSlice(Seconds(2), start=Seconds(11))
        ts2 = ts[sl]
        self.assertEqual(0, ts2.size)
        self.assertIsInstance(ts2, ArrayWithUnits)

    def test_time_slice_spanning_less_than_one_sample_returns_one_sample(self):
        arr = np.arange(10)
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        sl = TimeSlice(Milliseconds(100), start=Milliseconds(1500))
        ts2 = ts[sl]
        self.assertIsInstance(ts2, ArrayWithUnits)
        self.assertEqual(1, ts2.size)
        self.assertEqual(1, ts2[0])

    def test_time_slice_spanning_multiple_samples_returns_all_samples(self):
        arr = np.arange(10)
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        sl = TimeSlice(Milliseconds(2000), start=Milliseconds(1500))
        ts2 = ts[sl]
        self.assertIsInstance(ts2, ArrayWithUnits)
        self.assertEqual(3, ts2.size)
        self.assertTrue(np.all(np.arange(1, 4) == ts2))

    def test_frequency_and_duration_differ(self):
        arr = np.arange(10)
        freq = Seconds(1)
        duration = Seconds(2)
        ts = ArrayWithUnits(arr, [TimeDimension(freq, duration)])
        sl = TimeSlice(Seconds(2), start=Seconds(1))
        ts2 = ts[sl]
        self.assertIsInstance(ts2, ArrayWithUnits)
        self.assertEqual(3, ts2.size)
        self.assertTrue(np.all(np.arange(3) == ts2))

    def test_frequency_and_duration_differ2(self):
        arr = np.arange(10)
        freq = Seconds(1)
        duration = Seconds(3)
        ts = ArrayWithUnits(arr, [TimeDimension(freq, duration)])
        sl = TimeSlice(Seconds(2), start=Seconds(5))
        ts2 = ts[sl]
        self.assertIsInstance(ts2, ArrayWithUnits)
        self.assertEqual(4, ts2.size)
        self.assertTrue(np.all(np.arange(3, 7) == ts2))

    def test_frequency_and_duration_differ3(self):
        arr = np.arange(10)
        freq = Seconds(1)
        duration = Seconds(3)
        ts = ArrayWithUnits(arr, [TimeDimension(freq, duration)])
        sl = TimeSlice(Seconds(2), start=Seconds(6))
        ts2 = ts[sl]
        self.assertIsInstance(ts2, ArrayWithUnits)
        self.assertEqual(4, ts2.size)
        self.assertTrue(np.all(np.arange(4, 8) == ts2))

    def test_frequency_less_than_one(self):
        arr = np.arange(10)
        freq = Milliseconds(500)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        sl = TimeSlice(Seconds(2), start=Milliseconds(600))
        ts2 = ts[sl]
        self.assertIsInstance(ts2, ArrayWithUnits)
        self.assertEqual(5, ts2.size)
        self.assertTrue(np.all(np.arange(1, 6) == ts2))

    def test_frequency_less_than_one_freq_and_duration_differ(self):
        arr = np.arange(10)
        freq = Milliseconds(500)
        duration = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq, duration)])
        sl = TimeSlice(Seconds(3), start=Milliseconds(250))
        ts2 = ts[sl]
        self.assertIsInstance(ts2, ArrayWithUnits)
        self.assertEqual(7, ts2.size)
        self.assertTrue(np.all(np.arange(0, 7) == ts2))

    def test_frequency_less_than_one_freq_and_duration_differ2(self):
        arr = np.arange(10)
        freq = Milliseconds(500)
        duration = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq, duration)])
        sl = TimeSlice(Seconds(3), start=Milliseconds(1250))
        ts2 = ts[sl]
        self.assertIsInstance(ts2, ArrayWithUnits)
        self.assertEqual(8, ts2.size)
        self.assertTrue(np.all(np.arange(1, 9) == ts2))

    def test_duration_less_than_frequency(self):
        arr = np.arange(10)
        freq = Seconds(1)
        duration = Milliseconds(500)
        ts = ArrayWithUnits(arr, [TimeDimension(freq, duration)])
        sl = TimeSlice(Seconds(3), start=Milliseconds(1250))
        ts2 = ts[sl]
        self.assertIsInstance(ts2, ArrayWithUnits)
        self.assertEqual(4, ts2.size)
        self.assertTrue(np.all(np.arange(1, 5) == ts2))

    def test_can_get_entire_time_series_with_empty_slice(self):
        arr = np.arange(10)
        freq = Seconds(1)
        duration = Milliseconds(500)
        ts = ArrayWithUnits(arr, [TimeDimension(freq, duration)])
        ts2 = ts[:]
        self.assertIsInstance(ts2, ArrayWithUnits)
        self.assertTrue(np.all(np.arange(10) == ts2))

    def test_can_get_entire_time_series_with_empty_time_slice(self):
        arr = np.arange(10)
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        sl = TimeSlice()
        ts2 = ts[sl]
        self.assertEqual(10, len(ts2))

    def test_can_sum_2d_timeseries(self):
        arr = np.zeros((10, 3))
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq), IdentityDimension()])
        ts2 = ts.sum(axis=1)
        self.assertIsInstance(ts2, ArrayWithUnits)
        self.assertEqual(1, len(ts2.dimensions))
        self.assertEqual(freq, ts2.dimensions[0].frequency)
        self.assertEqual(freq, ts2.dimensions[0].duration)

    def test_span_freq_and_duration_equal(self):
        arr = np.arange(10)
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        self.assertEqual(TimeSlice(Seconds(10)), ts.dimensions[0].span)

    def test_span_duration_greater_than_frequency(self):
        arr = np.arange(10)
        freq = Seconds(1)
        duration = Milliseconds(2500)
        ts = ArrayWithUnits(arr, [TimeDimension(freq, duration)])
        self.assertEqual(TimeSlice(Milliseconds(11500)), ts.dimensions[0].span)

    def test_span_duration_less_than_frequency(self):
        arr = np.arange(10)
        freq = Seconds(1)
        duration = Milliseconds(500)
        ts = ArrayWithUnits(arr, [TimeDimension(freq, duration)])
        self.assertEqual(TimeSlice(Milliseconds(9500)), ts.dimensions[0].span)

    def test_duration_in_seconds_half_second(self):
        arr = np.arange(10)
        freq = Milliseconds(500)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        self.assertEqual(0.5, ts.dimensions[0].duration_in_seconds)

    def test_duration_in_seconds_two_seconds(self):
        arr = np.arange(10)
        freq = Seconds(2)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        self.assertEqual(2, ts.dimensions[0].duration_in_seconds)

    def test_samplerate_one_per_second(self):
        arr = np.arange(10)
        freq = Seconds(1)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        self.assertEqual(1, ts.dimensions[0].samples_per_second)

    def test_samplerate_two_per_second(self):
        arr = np.arange(10)
        freq = Milliseconds(500)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        self.assertEqual(2, ts.dimensions[0].samples_per_second)

    def test_samplerate_three_per_second(self):
        arr = np.arange(10)
        freq = Milliseconds(333)
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        self.assertEqual(3, int(ts.dimensions[0].samples_per_second))

    def test_samplerate_audio(self):
        arr = np.arange(10)
        freq = Picoseconds(int(1e12)) / 44100.
        ts = ArrayWithUnits(arr, [TimeDimension(freq)])
        self.assertEqual(44100, int(ts.dimensions[0].samples_per_second))

    def test_concatenation_with_differing_freqs_and_durations_raises(self):
        ts = ArrayWithUnits(
                np.arange(10),
                [TimeDimension(Seconds(1), Seconds(2))])
        ts2 = ArrayWithUnits(
                np.arange(10, 20),
                [TimeDimension(Seconds(1), Seconds(1))])
        self.assertRaises(ValueError, lambda: ts.concatenate(ts2))

    def test_concatenation_with_matching_freqs_and_duration_results_in_crts(
            self):
        ts = ArrayWithUnits(
                np.ones((10, 3)),
                [TimeDimension(Seconds(1), Seconds(2)), IdentityDimension()])
        ts2 = ArrayWithUnits(
                np.ones((13, 3)),
                [TimeDimension(Seconds(1), Seconds(2)), IdentityDimension()])
        result = ts.concatenate(ts2)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((23, 3), result.shape)

    def test_concat_with_differing_freqs(self):
        ts = ArrayWithUnits(
                np.ones((10, 3)),
                [TimeDimension(Seconds(2), Seconds(2)), IdentityDimension()])
        ts2 = ArrayWithUnits(
                np.ones((13, 3)),
                [TimeDimension(Seconds(1), Seconds(2)), IdentityDimension()])
        self.assertRaises(
                ValueError, lambda: ArrayWithUnits.concat([ts, ts2]))

    def test_concat_with_differing_durations(self):
        td1 = TimeDimension(Seconds(1), Seconds(2))
        ts1 = ArrayWithUnits(np.ones((10, 3)), [td1, IdentityDimension()])
        td2 = TimeDimension(Seconds(1), Seconds(3))
        ts2 = ArrayWithUnits(np.ones((13, 3)), [td2, IdentityDimension()])
        self.assertRaises(
                ValueError, lambda: ArrayWithUnits.concat([ts1, ts2]))

    def test_concat_along_first_axis(self):
        td1 = TimeDimension(Seconds(1), Seconds(2))
        ts1 = ArrayWithUnits(np.ones((10, 3)), [td1, IdentityDimension()])
        td2 = TimeDimension(Seconds(1), Seconds(2))
        ts2 = ArrayWithUnits(np.ones((13, 3)), [td2, IdentityDimension()])
        result = ArrayWithUnits.concat([ts1, ts2])
        self.assertEqual((23, 3), result.shape)

    def test_concat_along_second_axis(self):
        td1 = TimeDimension(Seconds(1), Seconds(2))
        ts1 = ArrayWithUnits(np.ones((10, 3)), [td1, IdentityDimension()])
        td2 = TimeDimension(Seconds(1), Seconds(2))
        ts2 = ArrayWithUnits(np.ones((10, 5)), [td2, IdentityDimension()])
        result = ArrayWithUnits.concat([ts1, ts2], axis=1)
        self.assertEqual((10, 8), result.shape)
        self.assertIsInstance(result.dimensions[0], TimeDimension)
        self.assertIsInstance(result.dimensions[1], IdentityDimension)

    def test_sum_along_time_axis(self):
        td = TimeDimension(Seconds(1), Seconds(2))
        ts = ArrayWithUnits(np.ones((10, 3)), [td, IdentityDimension()])
        result = ts.sum(axis=0)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((3,), result.shape)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], IdentityDimension)

    def test_sum_along_second_axis(self):
        td = TimeDimension(Seconds(1), Seconds(2))
        ts = ArrayWithUnits(np.ones((10, 3)), [td, IdentityDimension()])
        result = ts.sum(axis=1)
        self.assertIsInstance(result, ArrayWithUnits)
        self.assertEqual((10,), result.shape)
        self.assertEqual(1, len(result.dimensions))
        self.assertIsInstance(result.dimensions[0], TimeDimension)



from zounds.timeseries import TimeDimension
import numpy as np
from basedimension import BaseDimensionEncoder, BaseDimensionDecoder
import re
import base64


class TimeDimensionEncoder(BaseDimensionEncoder):
    DTYPE_RE = re.compile(r'\[(?P<dtype>[^\]]+)\]')

    def __init__(self):
        super(TimeDimensionEncoder, self).__init__(TimeDimension)

    def _encode_timedelta(self, td):
        dtype = self.DTYPE_RE.search(str(td.dtype)).groupdict()['dtype']
        return base64.b64encode(td.astype(np.uint64).tostring()), dtype

    def dict(self, o):
        return dict(
                frequency=self._encode_timedelta(o.frequency),
                duration=self._encode_timedelta(o.duration),
                size=o.size)


class TimeDimensionDecoder(BaseDimensionDecoder):
    def __init__(self):
        super(TimeDimensionDecoder, self).__init__(TimeDimension)

    def _decode_timedelta(self, t):
        if isinstance(t, np.timedelta64):
            return t

        v = np.fromstring(base64.b64decode(t[0]), dtype=np.uint64)[0]
        s = t[1]
        return np.timedelta64(long(v), s)

    def kwargs(self, d):
        return dict(
            frequency=self._decode_timedelta(d['frequency']),
            duration=self._decode_timedelta(d['duration']),
            size=d['size'])



import unittest2
from random_samples import ReservoirSampler
from zounds.timeseries import TimeDimension, Seconds
from zounds.spectral import FrequencyDimension, FrequencyBand, LinearScale
from zounds.core import ArrayWithUnits, IdentityDimension
import numpy as np


class TestReservoirSampler(unittest2.TestCase):
    def test_can_wrap_samples(self):
        def wrapper(empty, orig):
            return ArrayWithUnits(
                    empty, [IdentityDimension()] + list(orig.dimensions[1:]))

        sampler = ReservoirSampler(
                nsamples=10,
                wrapper=wrapper)

        frequency_dimension = FrequencyDimension(
                LinearScale(FrequencyBand(100, 1000), 100))

        samples = ArrayWithUnits(
                np.ones((2, 10, 100)),
                [
                    TimeDimension(frequency=Seconds(10)),
                    TimeDimension(frequency=Seconds(1)),
                    frequency_dimension
                ])

        sampler._enqueue(samples, pusher=None)
        reservoir = sampler._r
        self.assertEqual((10, 10, 100), reservoir.shape)
        self.assertIsInstance(reservoir, ArrayWithUnits)
        self.assertEqual(reservoir.dimensions[0], IdentityDimension())
        self.assertEqual(reservoir.dimensions[1], samples.dimensions[1])
        self.assertEqual(reservoir.dimensions[2], samples.dimensions[2])

    def test_default_wrapper_is_identity_function(self):
        sampler = ReservoirSampler(nsamples=10)

        frequency_dimension = FrequencyDimension(
                LinearScale(FrequencyBand(100, 1000), 100))

        samples = ArrayWithUnits(
                np.ones((2, 10, 100)),
                [
                    TimeDimension(frequency=Seconds(10)),
                    TimeDimension(frequency=Seconds(1)),
                    frequency_dimension
                ])

        sampler._enqueue(samples, pusher=None)
        reservoir = sampler._r
        self.assertEqual((10, 10, 100), reservoir.shape)
        self.assertNotIsInstance(reservoir, ArrayWithUnits)
        self.assertIsInstance(reservoir, np.ndarray)

    def test_can_dequeue_when_reservoir_is_full(self):
        def wrapper(empty, orig):
            return ArrayWithUnits(
                    empty, [IdentityDimension()] + list(orig.dimensions[1:]))

        sampler = ReservoirSampler(
                nsamples=10,
                wrapper=wrapper)

        frequency_dimension = FrequencyDimension(
                LinearScale(FrequencyBand(100, 1000), 100))

        samples = ArrayWithUnits(
                np.ones((10, 10, 100)),
                [
                    TimeDimension(frequency=Seconds(10)),
                    TimeDimension(frequency=Seconds(1)),
                    frequency_dimension
                ])

        sampler._enqueue(samples, pusher=None)
        reservoir = sampler._dequeue()

        self.assertEqual((10, 10, 100), reservoir.shape)
        self.assertIsInstance(reservoir, ArrayWithUnits)
        self.assertEqual(reservoir.dimensions[0], IdentityDimension())
        self.assertEqual(reservoir.dimensions[1], samples.dimensions[1])
        self.assertEqual(reservoir.dimensions[2], samples.dimensions[2])

    def test_can_dequeue_when_reservoir_is_partially_full(self):
        def wrapper(empty, orig):
            return ArrayWithUnits(
                    empty, [IdentityDimension()] + list(orig.dimensions[1:]))

        sampler = ReservoirSampler(
                nsamples=10,
                wrapper=wrapper)

        frequency_dimension = FrequencyDimension(
                LinearScale(FrequencyBand(100, 1000), 100))

        samples = ArrayWithUnits(
                np.ones((4, 10, 100)),
                [
                    TimeDimension(frequency=Seconds(10)),
                    TimeDimension(frequency=Seconds(1)),
                    frequency_dimension
                ])

        sampler._enqueue(samples, pusher=None)
        reservoir = sampler._dequeue()

        self.assertEqual((4, 10, 100), reservoir.shape)
        self.assertIsInstance(reservoir, ArrayWithUnits)
        self.assertEqual(reservoir.dimensions[0], IdentityDimension())
        self.assertEqual(reservoir.dimensions[1], samples.dimensions[1])
        self.assertEqual(reservoir.dimensions[2], samples.dimensions[2])



import featureflow as ff
import unittest2
import numpy as np

from zounds.util import simple_in_memory_settings
from preprocess import UnitNorm, PreprocessingPipeline
from learn import KMeans
from zounds.timeseries import Seconds, TimeDimension
from zounds.spectral import FrequencyBand, LinearScale, FrequencyDimension
from zounds.core import ArrayWithUnits, IdentityDimension


class UnitNormTests(unittest2.TestCase):
    def get_model(self):
        @simple_in_memory_settings
        class Model(ff.BaseModel):
            unitnorm = ff.PickleFeature(
                    UnitNorm,
                    store=False)

            kmeans = ff.PickleFeature(
                    KMeans,
                    centroids=10,
                    needs=unitnorm,
                    store=False)

            pipeline = ff.PickleFeature(
                    PreprocessingPipeline,
                    needs=(unitnorm, kmeans),
                    store=True)

        return Model

    def test_forward_transform_returns_array_with_units_where_possible(self):
        # train the model on random data
        training = np.random.random_sample((100, 30))
        Model = self.get_model()
        _id = Model.process(unitnorm=training)
        model = Model(_id)

        # create a time-frequency representation
        scale = LinearScale(FrequencyBand(20, 20000), 30)
        data = ArrayWithUnits(
                np.random.random_sample((10, 30)),
                [TimeDimension(Seconds(1)), FrequencyDimension(scale)])

        # do a forward pass
        transformed = model.pipeline.transform(data).data

        self.assertIsInstance(transformed, ArrayWithUnits)
        self.assertEqual(2, len(transformed.dimensions))
        self.assertIsInstance(transformed.dimensions[0], TimeDimension)
        self.assertIsInstance(transformed.dimensions[1], IdentityDimension)

    def invert_and_assert_class(self, data):
        training = np.random.random_sample((100, 30))
        Model = self.get_model()
        _id = Model.process(unitnorm=training)
        model = Model(_id)
        transformed = model.pipeline.transform(data)
        inverted = transformed.inverse_transform()
        self.assertEqual(data.__class__, inverted.__class__)
        return inverted

    def test_inversion_returns_array(self):
        data = np.random.random_sample((10, 30))
        self.invert_and_assert_class(data)

    def test_inversion_returns_time_series(self):
        data = np.random.random_sample((33, 30))
        ts = ArrayWithUnits(
                data, [TimeDimension(Seconds(1)), IdentityDimension()])
        inverted = self.invert_and_assert_class(ts)
        self.assertEqual(Seconds(1), inverted.dimensions[0].frequency)

    def test_inversion_returns_time_frequency_representation(self):
        data = np.random.random_sample((33, 30))
        scale = LinearScale(FrequencyBand(20, 20000), 30)
        tf = ArrayWithUnits(data, [
            TimeDimension(Seconds(1), Seconds(2)),
            FrequencyDimension(scale)])
        inverted = self.invert_and_assert_class(tf)
        self.assertEqual(Seconds(1), inverted.dimensions[0].frequency)
        self.assertEqual(Seconds(2), inverted.dimensions[0].duration)
        self.assertEqual(scale, inverted.dimensions[1].scale)

    def test_can_easily_get_at_codebook(self):
        # KLUDGE: This doesn't belong here, but putting it here is convenient
        # right now
        training = np.random.random_sample((100, 30))
        Model = self.get_model()
        _id = Model.process(unitnorm=training)
        model = Model(_id)
        self.assertIsInstance(model.pipeline[-1].codebook, np.ndarray)



from featureflow import Node, Aggregator
import numpy as np
from bisect import bisect_left
from zounds.timeseries import TimeSlice
from zounds.nputil import packed_hamming_distance
import featureflow as ff
from zounds.timeseries import VariableRateTimeSeriesFeature
from zounds.persistence import \
    ArrayWithUnitsFeature, PackedArrayWithUnitsEncoder, ArrayWithUnitsEncoder


class Contiguous(Node):
    def __init__(self, needs=None):
        super(Contiguous, self).__init__(needs=needs)

    def _process(self, data):
        _id, data = data
        yield data


class Offsets(Aggregator, Node):
    def __init__(self, needs=None):
        super(Offsets, self).__init__(needs=needs)
        self._cache = ([], [])
        self._offset = 0

    def _enqueue(self, data, pusher):
        _id, data = data
        self._cache[0].append(_id)
        self._cache[1].append(self._offset)
        self._offset += len(data)


class SearchResults(object):
    def __init__(self, query, offsets, indices, time_slice_builder):
        super(SearchResults, self).__init__()
        self.query = query
        self._indices = indices
        self._time_slice_builder = time_slice_builder
        self._offsets = offsets

    @staticmethod
    def _bisect(l, x):
        i = bisect_left(l, x)
        if i == len(l):
            return len(l) - 1
        if x < l[i]:
            i -= 1
        return i

    def __iter__(self):
        _ids, positions = self._offsets
        for i in self._indices:
            start_index = self._bisect(positions, i)
            diff = i - positions[start_index]
            _id = _ids[start_index]
            yield _id, self._time_slice_builder.build(_id, diff)


class Scorer(object):
    def __init__(self, index):
        super(Scorer, self).__init__()
        self.contiguous = index.contiguous

    def decode_query(self, binary_query):
        return np.fromstring(binary_query, dtype=self.contiguous.dtype)

    def score(self, query):
        raise NotImplementedError()


class HammingDistanceScorer(Scorer):
    def __init__(self, index):
        super(HammingDistanceScorer, self).__init__(index)

    def score(self, query):
        return np.logical_xor(query, self.contiguous).sum(axis=1)


class PackedHammingDistanceScorer(Scorer):
    def __init__(self, index):
        super(PackedHammingDistanceScorer, self).__init__(index)

    def score(self, query):
        return packed_hamming_distance(
                query.view(np.uint64), self.contiguous.view(np.uint64))


class TimeSliceBuilder(object):
    def __init__(self, index):
        super(TimeSliceBuilder, self).__init__()
        self.offsets = index.offsets
        self.contiguous = index.contiguous

    def build(self, _id, diff):
        raise NotImplementedError()


class ConstantRateTimeSliceBuilder(TimeSliceBuilder):
    def __init__(self, index):
        super(ConstantRateTimeSliceBuilder, self).__init__(index)

    def build(self, _id, diff):
        time_dim = self.contiguous.dimensions[0]
        start_time = time_dim.frequency * diff
        duration = time_dim.duration
        return TimeSlice(duration, start_time)


class VariableRateTimeSliceBuilder(TimeSliceBuilder):
    def __init__(self, index, feature_func):
        super(VariableRateTimeSliceBuilder, self).__init__(index)
        self.feature_func = feature_func

    def build(self, _id, diff):
        slices = list(self.feature_func(_id))
        return slices[diff]


class Search(object):
    def __init__(self, index, scorer, time_slice_builder):
        super(Search, self).__init__()
        self.offsets = index.offsets
        self.time_slice_builder = time_slice_builder
        self.scorer = scorer

    def decode_query(self, binary_query):
        return self.scorer.decode_query(binary_query)

    def search(self, query, n_results=5):
        scores = self.scorer.score(query)
        indices = np.argsort(scores)[:n_results]
        return SearchResults(query, self.offsets, indices,
                             self.time_slice_builder)


def hamming_index(document, feature, packed=True):
    if isinstance(feature, ArrayWithUnitsFeature):
        iterator = ((d._id, feature(_id=d._id, persistence=document)) \
                    for d in document)
        constant_rate = True
        feat = ArrayWithUnitsFeature
        encoder = PackedArrayWithUnitsEncoder \
            if packed else ArrayWithUnitsEncoder
    elif isinstance(feature, VariableRateTimeSeriesFeature):
        iterator = ((d._id, feature(_id=d._id, persistence=document).slicedata) \
                    for d in document)
        constant_rate = False
        feat = ff.NumpyFeature
        encoder = ff.PackedNumpyEncoder if packed else ff.NumpyEncoder
    else:
        raise ValueError(
                'feature must be either constant or variable rate timeseries')

    class Index(ff.BaseModel):
        codes = ff.Feature(
                ff.IteratorNode,
                store=False)

        contiguous = feat(
                Contiguous,
                needs=codes,
                encoder=encoder,
                store=True)

        offsets = ff.PickleFeature(
                Offsets,
                needs=codes,
                store=True)

        def __init__(self):
            super(Index, self).__init__()
            scorer = PackedHammingDistanceScorer \
                if packed else HammingDistanceScorer
            self._scorer = scorer(self)
            if constant_rate:
                self._slice_builder = ConstantRateTimeSliceBuilder(self)
            else:
                self._slice_builder = VariableRateTimeSliceBuilder(
                        self,
                        lambda x: feature(_id=x, persistence=document).slices)
            self._search = Search(
                    self,
                    scorer=self._scorer,
                    time_slice_builder=self._slice_builder)

        def decode_query(self, binary_query):
            return self._search.decode_query(binary_query)

        def random_query(self):
            query_index = np.random.randint(0, len(self.contiguous))
            return self.contiguous[query_index]

        def random_search(self, n_results=5):
            return self.search(self.random_query(), n_results=n_results)

        def search(self, query, n_results=5):
            return self._search.search(query, n_results=n_results)

        @classmethod
        def build(cls):
            cls.process(codes=iterator)

    return Index



import numpy as np
import featureflow as ff


def sigmoid(a):
    return 1. / (1 + np.exp(-a))


def stochastic_binary(a):
    return a > np.random.random_sample(a.shape)


def simple_settings(cls):
    """
    Create sane default persistence settings for learning pipelines
    :param cls: The class to decorate
    """

    class Settings(ff.PersistenceSettings):
        _id = cls.__name__
        id_provider = ff.StaticIdProvider(_id)
        key_builder = ff.StringDelimitedKeyBuilder()
        database = ff.FileSystemDatabase(
                path=_id, key_builder=key_builder, createdirs=True)

    class Model(cls, Settings):
        pass

    return Model



import datetime
from zounds.persistence import ArrayWithUnitsFeature
from zounds.timeseries import Seconds, Picoseconds, TimeSlice, AudioSamples
from zounds.segment import TimeSliceFeature
from zounds.index import SearchResults
from zounds.soundfile import OggVorbisFeature
from soundfile import SoundFile
from contentrange import ContentRange
import numpy as np
from featureflow import Decoder
import matplotlib
import json
import base64

matplotlib.use('Agg')
from matplotlib import pyplot as plt
from io import BytesIO


class TempResult(object):
    def __init__(
            self,
            data,
            content_type,
            is_partial=False,
            content_range=None):
        self.content_range = content_range
        self.data = data
        self.content_type = content_type
        self.timestamp = datetime.datetime.utcnow()
        self.is_partial = is_partial


class DefaultSerializer(object):
    def __init__(self, content_type):
        super(DefaultSerializer, self).__init__()
        self._content_type = content_type

    def matches(self, context):
        if context.feature is None:
            return False
        return context.feature.encoder.content_type == self._content_type

    @property
    def content_type(self):
        return self._content_type

    def serialize(self, context):
        document = context.document
        feature = context.feature
        slce = context.slce
        flo = feature(_id=document._id, decoder=Decoder(), persistence=document)
        if slce.start:
            flo.seek(slce.start)
        if slce.stop:
            value = flo.read(slce.stop - slce.start)
        else:
            value = flo.read()
        key = document.key_builder.build(
                document._id, feature.key, feature.version)
        total = document.database.size(key)
        return TempResult(
                value,
                self.content_type,
                is_partial=slce.start is not None or slce.stop is not None,
                content_range=ContentRange.from_slice(slce, total))


class AudioSamplesSerializer(object):
    def __init__(self):
        super(AudioSamplesSerializer, self).__init__()

    def matches(self, context):
        return isinstance(context.value, AudioSamples)

    @property
    def content_type(self):
        return 'audio/ogg'

    def serialize(self, context):
        bio = BytesIO()
        samples = context.value
        with SoundFile(
                bio,
                mode='w',
                samplerate=samples.samples_per_second,
                channels=samples.channels,
                format='OGG',
                subtype='VORBIS') as sf:
            for i in xrange(0, len(samples), samples.samples_per_second):
                sf.write(samples[i: i + samples.samples_per_second])
        bio.seek(0)
        return TempResult(bio.read(), 'audio/ogg')


class OggVorbisSerializer(object):
    """
    Serializer capable of handling range requests against ogg vorbis files
    """

    def __init__(self):
        super(OggVorbisSerializer, self).__init__()

    def matches(self, context):
        if context.feature is None:
            return False
        return \
            isinstance(context.feature, OggVorbisFeature) \
            and isinstance(context.slce, TimeSlice)

    @property
    def content_type(self):
        return 'audio/ogg'

    def serialize(self, context):
        feature = context.feature
        document = context.document
        slce = context.slce
        wrapper = feature(_id=document._id, persistence=document)
        samples = wrapper[slce]
        bio = BytesIO()
        with SoundFile(
                bio,
                mode='w',
                samplerate=wrapper.samplerate,
                channels=wrapper.channels,
                format='OGG',
                subtype='VORBIS') as sf:
            sf.write(samples)
        bio.seek(0)
        content_range = ContentRange.from_timeslice(
                slce, Picoseconds(int(1e12 * wrapper.duration_seconds)))
        return TempResult(
                bio.read(),
                'audio/ogg',
                is_partial=slce != TimeSlice(),
                content_range=content_range)


def generate_image(data, is_partial=False, content_range=None):
    fig = plt.figure()
    if len(data.shape) == 1:
        plt.plot(data)
    elif len(data.shape) == 2:
        data = np.abs(np.asarray(data))
        mat = plt.matshow(np.rot90(data), cmap=plt.cm.viridis)
        mat.axes.get_xaxis().set_visible(False)
        mat.axes.get_yaxis().set_visible(False)
    else:
        raise ValueError('cannot handle dimensions > 2')
    bio = BytesIO()
    plt.savefig(bio, bbox_inches='tight', pad_inches=0, format='png')
    bio.seek(0)
    fig.clf()
    plt.close('all')
    return TempResult(
            bio.read(),
            'image/png',
            is_partial=is_partial,
            content_range=content_range)


class ConstantRateTimeSeriesSerializer(object):
    def __init__(self):
        super(ConstantRateTimeSeriesSerializer, self).__init__()

    def matches(self, context):
        return \
            isinstance(context.feature, ArrayWithUnitsFeature) \
            and isinstance(context.slce, TimeSlice)

    @property
    def content_type(self):
        return 'image/png'

    def serialize(self, context):
        feature = context.feature
        document = context.document
        data = feature(_id=document._id, persistence=document)
        sliced_data = data[context.slce]
        td = data.dimensions[0]
        content_range = ContentRange.from_timeslice(
                context.slce, td.end)
        return generate_image(
                sliced_data,
                is_partial=True,
                content_range=content_range)


class NumpySerializer(object):
    def __init__(self):
        super(NumpySerializer, self).__init__()

    def matches(self, context):
        if context.document is not None \
                and isinstance(context.feature, ArrayWithUnitsFeature):
            return True

        return \
            isinstance(context.value, np.ndarray) \
            and len(context.value.shape) in (1, 2)

    @property
    def content_type(self):
        return 'image/png'

    def serialize(self, context):
        feature = context.feature
        document = context.document
        value = context.value
        if value is None:
            data = feature(_id=document._id, persistence=document)
        else:
            data = value
        return generate_image(data)


class AudioSliceSerializer(object):
    def __init__(
            self,
            content_type,
            visualization_feature,
            audio_feature,
            path_builder):
        self.audio_feature = audio_feature
        self.visualization_feature = visualization_feature
        self._content_type = content_type
        self._path_builder = path_builder

    @property
    def content_type(self):
        return self._content_type

    def _seconds(self, ts):
        return {
            'start': ts.start / Seconds(1),
            'duration': ts.duration / Seconds(1)
        }

    def _result(self, ts, _id):
        return {
            'audio': self._path_builder(_id, self.audio_feature.key),
            'visualization': self._path_builder(
                    _id, self.visualization_feature.key),
            'slice': self._seconds(ts)
        }

    def iter_results(self, context):
        raise NotImplementedError()

    def additional_data(self, context):
        return dict()

    def serialize(self, context):
        results = map(lambda x: self._result(*x), self.iter_results(context))
        output = {'results': results}
        output.update(self.additional_data(context))
        return TempResult(json.dumps(output), self.content_type)


class OnsetsSerializer(AudioSliceSerializer):
    def __init__(self, visualization_feature, audio_feature, path_builder):
        super(OnsetsSerializer, self).__init__(
                'application/vnd.zounds.onsets+json',
                visualization_feature,
                audio_feature,
                path_builder)

    def matches(self, context):
        return isinstance(context.feature, TimeSliceFeature)

    def iter_results(self, context):
        for ts in context.value:
            yield ts, context.document._id


class SearchResultsSerializer(AudioSliceSerializer):
    def __init__(self, visualization_feature, audio_feature, path_builder):
        super(SearchResultsSerializer, self).__init__(
                'application/vnd.zounds.searchresults+json',
                visualization_feature,
                audio_feature,
                path_builder)

    def matches(self, context):
        return isinstance(context.value, SearchResults)

    def additional_data(self, context):
        return {'query': base64.b64encode(context.value.query)}

    def iter_results(self, context):
        for _id, ts in context.value:
            yield ts, _id



import unittest2
from tfrepresentation import FrequencyDimension
from frequencyscale import FrequencyBand, LinearScale, LogScale


class FrequencyDimensionTests(unittest2.TestCase):
    def test_equal(self):
        fd1 = FrequencyDimension(LinearScale(FrequencyBand(20, 10000), 100))
        fd2 = FrequencyDimension(LinearScale(FrequencyBand(20, 10000), 100))
        self.assertEqual(fd1, fd2)

    def test_not_equal(self):
        fd1 = FrequencyDimension(LinearScale(FrequencyBand(20, 10000), 100))
        fd2 = FrequencyDimension(LogScale(FrequencyBand(20, 10000), 100))
        self.assertNotEqual(fd1, fd2)



import unittest2
import numpy as np
from weighting import AWeighting
from frequencyscale import LinearScale, FrequencyBand
from tfrepresentation import FrequencyDimension
from zounds.timeseries import Seconds, TimeDimension
from zounds.core import ArrayWithUnits, IdentityDimension


class WeightingTests(unittest2.TestCase):
    def test_cannot_multiply_when_array_does_not_have_expected_dimensions(self):
        td = TimeDimension(Seconds(1), Seconds(1))
        tf = ArrayWithUnits(np.ones((90, 100)), [td, IdentityDimension()])
        weighting = AWeighting()
        self.assertRaises(ValueError, lambda: tf * weighting)

    def test_can_get_weights_from_tf_representation(self):
        td = TimeDimension(Seconds(1), Seconds(1))
        fd = FrequencyDimension(LinearScale(FrequencyBand(20, 22050), 100))
        tf = ArrayWithUnits(np.ones((90, 100)), [td, fd])
        weighting = AWeighting()
        weights = weighting.weights(tf)
        self.assertEqual((100,), weights.shape)

    def test_can_get_weights_from_scale(self):
        scale = LinearScale(FrequencyBand(20, 22050), 100)
        weighting = AWeighting()
        weights = weighting.weights(scale)
        self.assertEqual((100,), weights.shape)



from featureflow import Node, Aggregator
import os
from soundfile import SoundFile
import requests
import json
import io
from urlparse import urlparse


class AudioMetaData(object):
    def __init__(
            self,
            uri=None,
            samplerate=None,
            channels=None,
            licensing=None,
            description=None,
            tags=None):
        super(AudioMetaData, self).__init__()
        self.uri = uri
        self.samplerate = samplerate
        self.channels = channels
        self.licensing = licensing
        self.description = description
        self.tags = tags

    def __eq__(self, other):
        return self.uri == other.uri \
               and self.samplerate == other.samplerate \
               and self.channels == other.channels \
               and self.licensing == other.licensing \
               and self.description == other.description \
               and self.tags == other.tags

    def __repr__(self):
        return self.__dict__.__str__()

    def __str__(self):
        return self.__repr__()


class AudioMetaDataEncoder(Aggregator, Node):
    content_type = 'application/json'

    def __init__(self, needs=None):
        super(AudioMetaDataEncoder, self).__init__(needs=needs)

    def _uri(self, uri):
        if isinstance(uri, requests.Request):
            return uri.uri
        elif isinstance(uri, io.BytesIO):
            return None
        else:
            return uri

    def _process(self, data):
        print data
        yield json.dumps({
            'uri': self._uri(data.uri),
            'samplerate': data.samplerate,
            'channels': data.channels,
            'licensing': data.licensing,
            'description': data.description,
            'tags': data.tags
        })


class FreesoundOrgConfig(object):
    def __init__(self, api_key):
        super(FreesoundOrgConfig, self).__init__()
        self.api_key = api_key

    def request(self, _id):
        uri = 'http://freesound.org/apiv2/sounds/{_id}/'.format(_id=_id)
        params = {'token': self.api_key}
        metadata = requests.get(uri, params=params).json()
        request = requests.Request(
                method='GET',
                url=metadata['previews']['preview-hq-ogg'],
                params=params)
        return AudioMetaData(
                uri=request,
                samplerate=metadata['samplerate'],
                channels=metadata['channels'],
                licensing=metadata['license'],
                description=metadata['description'],
                tags=metadata['tags'])


class MetaData(Node):
    def __init__(self, needs=None):
        super(MetaData, self).__init__(needs=needs)

    @staticmethod
    def _is_url(s):
        if not isinstance(s, str):
            return False
        parsed = urlparse(s)
        return parsed.scheme and parsed.netloc

    @staticmethod
    def _is_local_file(s):
        try:
            return os.path.exists(s)
        except TypeError:
            return False

    @staticmethod
    def _is_file(s):
        try:
            s.tell()
            return True
        except AttributeError:
            return False

    def _process(self, data):
        if isinstance(data, AudioMetaData):
            yield data
        elif self._is_url(data):
            req = requests.Request(
                    method='GET',
                    url=data,
                    headers={'Range': 'bytes=0-'})
            yield AudioMetaData(uri=req)
        elif isinstance(data, requests.Request):
            if 'range' not in data.headers:
                data.headers['range'] = 'bytes=0-'
            yield AudioMetaData(uri=data)
        elif self._is_local_file(data) or self._is_file(data):
            sf = SoundFile(data)
            yield AudioMetaData(
                    uri=data,
                    samplerate=sf.samplerate,
                    channels=sf.channels)
        else:
            yield AudioMetaData(uri=data)



import unittest2
from zounds.core import ArrayWithUnits, IdentityDimension
from zounds.timeseries import \
    TimeDimension, Seconds, Milliseconds, AudioSamples, SR11025
from zounds.spectral import FrequencyDimension, FrequencyScale, FrequencyBand
from arraywithunits import \
    ArrayWithUnitsEncoder, ArrayWithUnitsDecoder, PackedArrayWithUnitsEncoder
import numpy as np
from io import BytesIO


class ArrayWithUnitsFeatureTests(unittest2.TestCase):
    def _roundtrip(self, arr, encoder=None):
        encoder = encoder or ArrayWithUnitsEncoder()
        decoder = ArrayWithUnitsDecoder()
        encoded = BytesIO(''.join(encoder._process(arr)))
        return decoder(encoded)

    def test_can_pack_bits(self):
        raw = np.random.binomial(1, 0.5, (100, 64))
        arr = ArrayWithUnits(
                raw, [TimeDimension(Seconds(1)), IdentityDimension()])
        decoded = self._roundtrip(arr, encoder=PackedArrayWithUnitsEncoder())
        self.assertIsInstance(decoded, ArrayWithUnits)
        self.assertEqual(2, len(decoded.dimensions))
        self.assertEqual((100, 8), decoded.shape)

    def test_can_round_trip_audio_samples(self):
        raw = np.random.random_sample(11025 * 10)
        arr = AudioSamples(raw, SR11025())
        decoded = self._roundtrip(arr)
        self.assertIsInstance(decoded, ArrayWithUnits)
        self.assertEqual(1, len(decoded.dimensions))
        td = decoded.dimensions[0]
        self.assertIsInstance(td, TimeDimension)
        np.testing.assert_allclose(decoded, raw)

    def test_can_round_trip_1d_identity_dimension(self):
        raw = np.arange(10)
        arr = ArrayWithUnits(raw, (IdentityDimension(),))
        decoded = self._roundtrip(arr)
        self.assertIsInstance(decoded, ArrayWithUnits)
        self.assertEqual(1, len(decoded.dimensions))
        d = decoded.dimensions[0]
        self.assertIsInstance(d, IdentityDimension)
        np.testing.assert_allclose(decoded, raw)

    def test_can_round_trip_2d_with_identity_dimension(self):
        raw = np.random.random_sample((10, 10))
        dim = TimeDimension(Seconds(1), Milliseconds(500))
        arr = ArrayWithUnits(raw, (IdentityDimension(), dim))
        decoded = self._roundtrip(arr)
        self.assertIsInstance(decoded, ArrayWithUnits)
        self.assertEqual(2, len(decoded.dimensions))
        idd = decoded.dimensions[0]
        self.assertIsInstance(idd, IdentityDimension)
        td = decoded.dimensions[1]
        self.assertIsInstance(td, TimeDimension)
        self.assertEqual(Seconds(1), td.frequency)
        self.assertEqual(Milliseconds(500), td.duration)

        np.testing.assert_allclose(decoded, raw)

    def test_can_round_trip_1d_constant_rate_time_series(self):
        dim = TimeDimension(Seconds(1), Milliseconds(500))
        raw = np.arange(10)
        ts = ArrayWithUnits(raw, (dim,))
        decoded = self._roundtrip(ts)
        self.assertIsInstance(decoded, ArrayWithUnits)
        self.assertEqual(1, len(decoded.dimensions))
        self.assertIsInstance(decoded.dimensions[0], TimeDimension)
        td = decoded.dimensions[0]
        self.assertEqual(Seconds(1), td.frequency)
        self.assertEqual(Milliseconds(500), td.duration)
        np.testing.assert_allclose(decoded, raw)

    def test_can_round_trip_2d_constant_rate_time_series(self):
        dim1 = TimeDimension(Seconds(1), Milliseconds(500))
        scale = FrequencyScale(FrequencyBand(20, 20000), 100)
        dim2 = FrequencyDimension(scale)
        raw = np.random.random_sample((10, 100))
        ts = ArrayWithUnits(raw, (dim1, dim2))
        decoded = self._roundtrip(ts)
        self.assertIsInstance(decoded, ArrayWithUnits)
        self.assertEqual(2, len(decoded.dimensions))
        self.assertIsInstance(decoded.dimensions[0], TimeDimension)
        td = decoded.dimensions[0]
        self.assertIsInstance(td, TimeDimension)
        self.assertEqual(Seconds(1), td.frequency)
        self.assertEqual(Milliseconds(500), td.duration)
        fd = decoded.dimensions[1]
        self.assertIsInstance(fd, FrequencyDimension)
        self.assertEqual(scale, fd.scale)
        np.testing.assert_allclose(decoded, raw)

    def test_can_round_trip_3d_constant_rate_time_series_with_frequency_dim(
            self):
        dim1 = TimeDimension(Seconds(2), Milliseconds(1000))
        dim2 = TimeDimension(Seconds(1), Milliseconds(500))
        scale = FrequencyScale(FrequencyBand(20, 20000), 100)
        dim3 = FrequencyDimension(scale)
        raw = np.random.random_sample((5, 2, 100))
        ts = ArrayWithUnits(raw, (dim1, dim2, dim3))

        decoded = self._roundtrip(ts)
        self.assertIsInstance(decoded, ArrayWithUnits)
        self.assertEqual(3, len(decoded.dimensions))

        td1 = decoded.dimensions[0]
        self.assertIsInstance(td1, TimeDimension)
        self.assertEqual(Seconds(2), td1.frequency)
        self.assertEqual(Milliseconds(1000), td1.duration)

        td2 = decoded.dimensions[1]
        self.assertIsInstance(td2, TimeDimension)
        self.assertEqual(Seconds(1), td2.frequency)
        self.assertEqual(Milliseconds(500), td2.duration)

        fd = decoded.dimensions[2]
        self.assertIsInstance(fd, FrequencyDimension)
        self.assertEqual(scale, fd.scale)
        np.testing.assert_allclose(decoded, raw)





from __future__ import division
import numpy as np
from scipy.fftpack import idct
from zounds.timeseries import audio_sample_rate, Seconds, AudioSamples
from zounds.spectral import DCTIV
from zounds.spectral.sliding_window import \
    IdentityWindowingFunc, OggVorbisWindowingFunc
from zounds.core import ArrayWithUnits, IdentityDimension


class ShortTimeTransformSynthesizer(object):
    def __init__(self):
        super(ShortTimeTransformSynthesizer, self).__init__()

    def _transform(self, frames):
        return frames

    def _windowing_function(self):
        return IdentityWindowingFunc()

    def _overlap_add(self, frames):
        # BUG: This code assumes there is a coefficient for every raw audio
        # sample, but this isn't the case for FFT, e.g.
        time_dim = frames.dimensions[0]
        sample_length_seconds = time_dim.duration_in_seconds / frames.shape[-1]
        samples_per_second = int(1 / sample_length_seconds)
        samplerate = audio_sample_rate(samples_per_second)
        windowsize = int(np.round(time_dim.duration / samplerate.frequency))
        hopsize = int(np.round(time_dim.frequency / samplerate.frequency))
        arr = np.zeros(int(time_dim.end / samplerate.frequency))
        for i, f in enumerate(frames):
            start = i * hopsize
            stop = start + windowsize
            l = len(arr[start:stop])
            arr[start:stop] += (self._windowing_function() * f[:l])
        return AudioSamples(arr, samplerate)

    def synthesize(self, frames):
        audio = self._transform(frames)
        ts = ArrayWithUnits(audio, [frames.dimensions[0], IdentityDimension()])
        return self._overlap_add(ts)


class FFTSynthesizer(ShortTimeTransformSynthesizer):
    def __init__(self):
        super(FFTSynthesizer, self).__init__()

    def _transform(self, frames):
        return np.fft.fftpack.ifft(frames)


class DCTSynthesizer(ShortTimeTransformSynthesizer):
    def __init__(self, windowing_func=IdentityWindowingFunc()):
        super(DCTSynthesizer, self).__init__()
        self.windowing_func = windowing_func

    def _windowing_function(self):
        return self.windowing_func

    def _transform(self, frames):
        return idct(frames, norm='ortho')


class DCTIVSynthesizer(ShortTimeTransformSynthesizer):
    """
    Perform the inverse of the DCTIV transform, which is the same as the forward
    transformation
    """

    def __init__(self, windowing_func=IdentityWindowingFunc()):
        super(DCTIVSynthesizer, self).__init__()
        self.windowing_func = windowing_func

    def _windowing_function(self):
        return self.windowing_func

    def _transform(self, frames):
        return list(DCTIV()._process(frames))[0]


class MDCTSynthesizer(ShortTimeTransformSynthesizer):
    def __init__(self):
        super(MDCTSynthesizer, self).__init__()

    def _windowing_function(self):
        return OggVorbisWindowingFunc()

    def _transform(self, frames):
        l = frames.shape[1]
        t = np.arange(0, 2 * l)
        f = np.arange(0, l)
        cpi = -1j * np.pi
        a = frames * np.exp(cpi * (f + 0.5) * (l + 1) / 2 / l)
        b = np.fft.fft(a, 2 * l)
        return np.sqrt(2 / l) * np.real(b * np.exp(cpi * t / 2 / l))


class SineSynthesizer(object):
    """
    Synthesize sine waves
    """

    def __init__(self, samplerate):
        super(SineSynthesizer, self).__init__()
        self.samplerate = samplerate

    def synthesize(self, duration, freqs_in_hz=[440.]):
        freqs = np.array(freqs_in_hz)
        scaling = 1 / len(freqs)
        sr = int(self.samplerate)
        cps = freqs / sr
        ts = (duration / Seconds(1)) * sr
        ranges = np.array([np.arange(0, ts * c, c) for c in cps])
        raw = (np.sin(ranges * (2 * np.pi)) * scaling).sum(axis=0)
        return AudioSamples(raw, self.samplerate)


class TickSynthesizer(object):
    """
    Synthesize short, percussive, periodic "ticks"
    """

    def __init__(self, samplerate):
        super(TickSynthesizer, self).__init__()
        self.samplerate = samplerate

    def synthesize(self, duration, tick_frequency):
        sr = self.samplerate.samples_per_second
        # create a short, tick sound
        tick = np.random.random_sample(int(sr * .1))
        tick *= np.linspace(1, 0, len(tick))
        # create silence
        samples = np.zeros(int(sr * (duration / Seconds(1))))
        ticks_per_second = Seconds(1) / tick_frequency
        # introduce periodic ticking sound
        step = int(sr // ticks_per_second)
        print 'STEP', step
        for i in xrange(0, len(samples), step):
            print 'STEP 2', i
            samples[i:i + len(tick)] = tick
        return AudioSamples(samples, self.samplerate)


class NoiseSynthesizer(object):
    """
    Synthesize white noise
    """

    def __init__(self, samplerate):
        super(NoiseSynthesizer, self).__init__()
        self.samplerate = samplerate

    def synthesize(self, duration):
        sr = self.samplerate.samples_per_second
        seconds = duration / Seconds(1)
        return AudioSamples(
                np.random.random_sample(int(sr * seconds)), self.samplerate)



import unittest2
import numpy as np
from duration import Seconds
from zounds.core import ArrayWithUnits
from timeseries import TimeDimension
from samplerate import \
    SampleRate, SR96000, SR48000, SR44100, SR22050, SR11025, audio_sample_rate, \
    HalfLapped


class SampleRateTests(unittest2.TestCase):

    def test_can_unpack_samplerate(self):
        sr = SampleRate(Seconds(1), Seconds(2))
        frequency, duration = sr
        self.assertEqual(Seconds(1), frequency)
        self.assertEqual(Seconds(2), duration)

    def test_can_unpack_audio_samplerate(self):
        sr = SR44100()
        frequency, duration = sr
        self.assertEqual(sr.frequency, frequency)
        self.assertEqual(sr.duration, duration)

    def test_discrete_samples_11025(self):
        sr = SR11025()
        ts = ArrayWithUnits(
                np.zeros(sr.samples_per_second), [TimeDimension(*sr)])
        hl = HalfLapped()
        freq, duration = hl.discrete_samples(ts)
        self.assertEqual(256, freq)
        self.assertEqual(512, duration)

    def test_discrete_samples_22050(self):
        sr = SR22050()
        ts = ArrayWithUnits(
                np.zeros(sr.samples_per_second), [TimeDimension(*sr)])
        hl = HalfLapped()
        freq, duration = hl.discrete_samples(ts)
        self.assertEqual(512, freq)
        self.assertEqual(1024, duration)

    def test_discrete_samples_44100(self):
        sr = SR44100()
        ts = ArrayWithUnits(
                np.zeros(sr.samples_per_second), [TimeDimension(*sr)])
        hl = HalfLapped()
        freq, duration = hl.discrete_samples(ts)
        self.assertEqual(1024, freq)
        self.assertEqual(2048, duration)

    def test_nyquist_22050(self):
        self.assertEqual(11025, SR22050().nyquist)

    def test_nyquist_44100(self):
        self.assertEqual(22050, SR44100().nyquist)

    def test_can_convert_to_int(self):
        self.assertEqual(22050, int(SR22050()))

    def test_raises_for_unknown_audio_samplerate(self):
        self.assertRaises(ValueError, lambda: audio_sample_rate(1))

    def test_sr_96000_frequency(self):
        self.assertEqual(96000, SR96000().samples_per_second)

    def test_get_96000_frequency(self):
        self.assertIsInstance(audio_sample_rate(96000), SR96000)

    def test_sr_48000_frequency(self):
        self.assertEqual(48000, SR48000().samples_per_second)

    def test_get_48000_frequency(self):
        self.assertIsInstance(audio_sample_rate(48000), SR48000)

    def test_sr_44100_frequency(self):
        self.assertEqual(44100, SR44100().samples_per_second)

    def test_get_44100_frequency(self):
        self.assertIsInstance(audio_sample_rate(44100), SR44100)

    def test_sr_22050_frequency(self):
        self.assertEqual(22050, SR22050().samples_per_second)

    def test_get_22050_freuency(self):
        self.assertIsInstance(audio_sample_rate(22050), SR22050)

    def test_sr_11025_frequency(self):
        self.assertEqual(11025, SR11025().samples_per_second)

    def test_get_11025_frequency(self):
        self.assertIsInstance(audio_sample_rate(11025), SR11025)

    def test_no_overlap(self):
        self.assertEqual(Seconds(0), SampleRate(Seconds(1), Seconds(1)).overlap)

    def test_some_overlap(self):
        self.assertEqual(Seconds(1), SampleRate(Seconds(1), Seconds(2)).overlap)

    def test_multiply_no_overlap_number(self):
        sr = SampleRate(Seconds(1), Seconds(1)) * 2
        self.assertEqual(Seconds(2), sr.frequency)
        self.assertEqual(Seconds(2), sr.duration)

    def test_multiply_some_overlap_number(self):
        sr = SampleRate(Seconds(1), Seconds(2)) * 2
        self.assertEqual(Seconds(2), sr.frequency)
        self.assertEqual(Seconds(3), sr.duration)

    def test_multiply_no_overlap_single_value(self):
        sr = SampleRate(Seconds(1), Seconds(1)) * (2,)
        self.assertEqual(Seconds(2), sr.frequency)
        self.assertEqual(Seconds(2), sr.duration)

    def test_multiply_some_overlap_single_value(self):
        sr = SampleRate(Seconds(1), Seconds(2)) * (2,)
        self.assertEqual(Seconds(2), sr.frequency)
        self.assertEqual(Seconds(3), sr.duration)

    def test_multiply_no_overlap_two_values(self):
        sr = SampleRate(Seconds(1), Seconds(1)) * (2, 4)
        self.assertEqual(Seconds(2), sr.frequency)
        self.assertEqual(Seconds(4), sr.duration)

    def test_multiply_some_overlap_two_values(self):
        sr = SampleRate(Seconds(1), Seconds(2)) * (2, 4)
        self.assertEqual(Seconds(2), sr.frequency)
        self.assertEqual(Seconds(5), sr.duration)



from index import \
    Contiguous, Offsets, HammingDistanceScorer, PackedHammingDistanceScorer, \
    ConstantRateTimeSliceBuilder, VariableRateTimeSliceBuilder, Search, \
    SearchResults, hamming_index



from sliding_window import \
    SlidingWindow, OggVorbisWindowingFunc, NDSlidingWindow, WindowingFunc

from spectral import \
    FFT, DCT, DCTIV, MDCT, BarkBands, Chroma, BFCC, SpectralCentroid, \
    SpectralFlatness

from tfrepresentation import FrequencyDimension

from weighting import AWeighting

from frequencyscale import LinearScale, LogScale, FrequencyBand, FrequencyScale



import numpy as np
import scipy
from featureflow import Node, NotEnoughData

from zounds.core import ArrayWithUnits
from zounds.nputil import sliding_window
from zounds.timeseries import TimeSlice


def oggvorbis(s):
    """
    This is taken from the ogg vorbis spec
    (http://xiph.org/vorbis/doc/Vorbis_I_spec.html)

    :param s: the total length of the window, in samples
    """
    try:
        s = np.arange(s)
    except TypeError:
        s = np.arange(s[0])

    i = np.sin((s + .5) / len(s) * np.pi) ** 2
    f = np.sin(.5 * np.pi * i)
    return f * (1. / f.max())


class WindowingFunc(object):

    def __init__(self, windowing_func=lambda size: np.ones(size)):
        super(WindowingFunc, self).__init__()
        self.windowing_func = windowing_func

    def _wdata(self, size):
        return self.windowing_func(size)

    def __numpy_ufunc__(self, *args, **kwargs):
        # KLUDGE: This seems really odd, but the mere presence of this
        # numpy-specific magic method seems to serve as a hint to to call
        # this instances __rmul__ implementation, instead of doing element-wise
        # multiplication, as per:
        # http://docs.scipy.org/doc/numpy/reference/arrays.classes.html#numpy.class.__numpy_ufunc__
        raise NotImplementedError()

    def __mul__(self, other):
        size = other.shape[-1]
        return self._wdata(size) * other

    def __rmul__(self, other):
        return self.__mul__(other)


class IdentityWindowingFunc(WindowingFunc):
    def __init__(self):
        super(IdentityWindowingFunc, self).__init__()


class OggVorbisWindowingFunc(WindowingFunc):
    def __init__(self):
        super(OggVorbisWindowingFunc, self).__init__(windowing_func=oggvorbis)


class SlidingWindow(Node):
    def __init__(self, wscheme, wfunc=None, padwith=0, needs=None):
        super(SlidingWindow, self).__init__(needs=needs)
        self._scheme = wscheme
        self._func = wfunc or IdentityWindowingFunc()
        self._padwith = padwith
        self._cache = None

    def _first_chunk(self, data):
        if self._padwith:
            padding = np.zeros(
                    (self._padwith,) + data.shape[1:], dtype=data.dtype)
            padding_ts = ArrayWithUnits(padding, data.dimensions)
            return padding_ts.concatenate(data)
        else:
            return data

    def _enqueue(self, data, pusher):
        if self._cache is None:
            self._cache = data
        else:
            self._cache = self._cache.concatenate(data)

    def _dequeue(self):

        duration = TimeSlice(self._scheme.duration)
        frequency = TimeSlice(self._scheme.frequency)

        leftover, arr = self._cache.sliding_window_with_leftovers(
                duration,
                frequency,
                dopad=self._finalized)

        if not arr.size:
            raise NotEnoughData()

        self._cache = leftover

        # BUG: Order matters here (try arr * self._func instead)
        # why does that statement result in __rmul__ being called for each
        # scalar value in arr?
        out = (self._func * arr) if self._func else arr

        return out


# KLUDGE: This extractor works when trying to get random patches over the whole
# database, but doesn't work if we were trying to get a sliding window over, say,
# the bark bands of a single sound.  For that to work, we'd have to implement
# enqueue and dequeue methods that keep leftovers around between calls.
#
# If we're processing bark bands of a single sound all at once, we expect the
# sliding window to move over the time dimension first, and then over the frequency
# dimension.  Since each incoming chunk is of inderteminate size, we're left with
# nonsense ouput. This problem could be addressed by transposing the incoming data,
# so that the frequency dimension is traversed first, and then the time dimension.
class NDSlidingWindow(Node):
    def __init__(self, windowsize=None, stepsize=None, needs=None):
        super(NDSlidingWindow, self).__init__(needs=needs)
        self._ws = windowsize
        self._ss = stepsize

    def _process(self, data):
        try:
            yield sliding_window(data, self._ws, self._ss)
        except ValueError:
            pass



import unittest2
import numpy as np
from synthesize import \
    ShortTimeTransformSynthesizer, SineSynthesizer, DCTSynthesizer
from zounds.timeseries import \
    SR22050, SR44100, HalfLapped, Seconds, TimeDimension, AudioSamples
from zounds.core import ArrayWithUnits
from zounds.spectral import FrequencyDimension, FrequencyBand, LinearScale


class SynthesizeTests(unittest2.TestCase):
    def test_has_correct_sample_rate(self):
        half_lapped = HalfLapped()
        # synth = ShortTimeTransformSynthesizer()
        synth = DCTSynthesizer()
        raw = np.zeros((100, 2048))
        band = FrequencyBand(0, SR44100().nyquist)
        scale = LinearScale(band, raw.shape[1])
        timeseries = ArrayWithUnits(
                raw, [TimeDimension(*half_lapped), FrequencyDimension(scale)])
        output = synth.synthesize(timeseries)
        self.assertIsInstance(output.samplerate, SR44100)
        self.assertIsInstance(output, AudioSamples)


class SineSynthesizerTests(unittest2.TestCase):

    def test_generates_correct_shape(self):
        ss = SineSynthesizer(SR22050())
        audio = ss.synthesize(Seconds(4), freqs_in_hz=[440.])
        self.assertEqual(1, len(audio.shape))

    def test_generates_correct_samplerate(self):
        ss = SineSynthesizer(SR44100())
        audio = ss.synthesize(Seconds(4), freqs_in_hz=[440.])
        self.assertEqual(SR44100(), audio.samplerate)

    def test_generates_correct_number_of_samples(self):
        samplerate = SR22050()
        duration = Seconds(1)
        ss = SineSynthesizer(samplerate)
        audio = ss.synthesize(duration, freqs_in_hz=[440.])
        expected_samples = int(samplerate) * int(duration / Seconds(1))
        self.assertEqual(expected_samples, len(audio))

    def test_can_create_audio_with_single_tone(self):
        ss = SineSynthesizer(SR22050())
        audio = ss.synthesize(Seconds(4), freqs_in_hz=[440.])
        fft = abs(np.fft.rfft(audio))
        self.assertEqual(1, (fft > 1).sum())

    def test_can_create_audio_with_multiple_tones(self):
        ss = SineSynthesizer(SR22050())
        audio = ss.synthesize(Seconds(4), freqs_in_hz=[440., 660.])
        fft = abs(np.fft.rfft(audio))
        self.assertEqual(2, (fft > 1).sum())



import re
from zounds.timeseries import Seconds, Picoseconds, TimeSlice


class RangeUnitUnsupportedException(Exception):
    """
    Raised when an HTTP range request is made with a unit not supported by this
    application
    """
    pass


class UnsatisfiableRangeRequestException(Exception):
    """
    Exception raised when an HTTP range request cannot be satisfied for a
    particular resource
    """
    pass


class RangeRequest(object):
    def __init__(self, range_header):
        self.range_header = range_header
        self.re = re.compile(
                r'^(?P<unit>[^=]+)=(?P<start>[^-]+)-(?P<stop>.*?)$')

    def time_slice(self, start, stop):
        start = float(start)
        try:
            stop = float(stop)
        except ValueError:
            stop = None
        duration = \
            None if stop is None else Picoseconds(int(1e12 * (stop - start)))
        start = Picoseconds(int(1e12 * start))
        return TimeSlice(duration, start=start)

    def byte_slice(self, start, stop):
        start = int(start)
        try:
            stop = int(stop)
        except ValueError:
            stop = None
        return slice(start, stop)

    def range(self):
        raw = self.range_header
        if not raw:
            return slice(None)

        m = self.re.match(raw)
        if not m:
            return slice(None)

        units = m.groupdict()['unit']
        start = m.groupdict()['start']
        stop = m.groupdict()['stop']

        if units == 'bytes':
            return self.byte_slice(start, stop)
        elif units == 'seconds':
            return self.time_slice(start, stop)
        else:
            raise RangeUnitUnsupportedException(units)


class ContentRange(object):
    def __init__(self, unit, start, total, stop=None):
        self.unit = unit
        self.total = total
        self.stop = stop
        self.start = start

    @staticmethod
    def from_timeslice(timeslice, total):
        one_second = Seconds(1)
        stop = None
        start = timeslice.start / one_second
        if timeslice.duration is not None:
            stop = start + (timeslice.duration / one_second)
        return ContentRange(
                'seconds',
                timeslice.start / one_second,
                total / one_second,
                stop)

    @staticmethod
    def from_slice(slce, total):
        return ContentRange('bytes', slce.start, total, slce.stop)

    def __str__(self):
        unit = self.unit
        start = self.start
        stop = self.stop or self.total
        total = self.total
        return '{unit} {start}-{stop}/{total}'.format(**locals())



from duration import \
    Hours, Minutes, Seconds, Milliseconds, Microseconds, Picoseconds

from audiosamples import AudioSamples

from samplerate import \
    SR11025, SR22050, SR44100, SR48000, SR96000, HalfLapped, audio_sample_rate,\
    Stride, SampleRate

from timeseries import TimeSlice, TimeDimension

from variablerate import VariableRateTimeSeries, VariableRateTimeSeriesFeature



import unittest2
import numpy as np
from preprocess import Multiply, PreprocessingPipeline
from zounds.util import simple_in_memory_settings
import featureflow as ff


class MultiplyTests(unittest2.TestCase):
    def get_model(self, factor):
        @simple_in_memory_settings
        class Model(ff.BaseModel):
            multiply = ff.PickleFeature(
                    Multiply,
                    factor=factor,
                    store=False)

            pipeline = ff.PickleFeature(
                    PreprocessingPipeline,
                    needs=(multiply,),
                    store=True)

        return Model

    def test_can_do_forward_transform_with_scalar(self):
        training = np.random.random_sample((100, 30))
        factor = 10
        Model = self.get_model(factor)
        _id = Model.process(multiply=training)
        model = Model(_id)
        data = np.random.random_sample((10, 30))
        transformed = model.pipeline.transform(data)
        np.testing.assert_allclose(data * factor, transformed.data)

    def test_can_do_forward_transform_with_array(self):
        training = np.random.random_sample((100, 30))
        factor = np.random.random_sample(30)
        Model = self.get_model(factor)
        _id = Model.process(multiply=training)
        model = Model(_id)
        data = np.random.random_sample((10, 30))
        transformed = model.pipeline.transform(data)
        np.testing.assert_allclose(data * factor, transformed.data)

    def test_raises_if_shapes_do_not_match(self):
        training = np.random.random_sample((100, 30))
        factor = np.random.random_sample(3)
        Model = self.get_model(factor)
        self.assertRaises(ValueError, lambda: Model.process(multiply=training))

    def test_can_do_forward_and_backward_transform_with_scalar(self):
        training = np.random.random_sample((100, 30))
        factor = 10
        Model = self.get_model(factor)
        _id = Model.process(multiply=training)
        model = Model(_id)
        data = np.random.random_sample((10, 30))
        transformed = model.pipeline.transform(data)
        recon = transformed.inverse_transform()
        np.testing.assert_allclose(data, recon)

    def test_can_do_forward_and_backward_transform_with_array(self):
        training = np.random.random_sample((100, 30))
        factor = np.random.random_sample(30)
        Model = self.get_model(factor)
        _id = Model.process(multiply=training)
        model = Model(_id)
        data = np.random.random_sample((10, 30))
        transformed = model.pipeline.transform(data)
        recon = transformed.inverse_transform()
        np.testing.assert_allclose(data, recon)



import unittest2
import featureflow as ff
import numpy as np

from zounds.util import simple_in_memory_settings
from preprocess import Flatten, PreprocessingPipeline


class FlattenTests(unittest2.TestCase):
    def do_setup(self, shape):
        @simple_in_memory_settings
        class Model(ff.BaseModel):
            flattened = ff.PickleFeature(
                    Flatten,
                    store=False)

            pipeline = ff.PickleFeature(
                    PreprocessingPipeline,
                    needs=(flattened,),
                    store=True)

        training = np.random.random_sample(shape)
        _id = Model.process(flattened=training)
        model = Model(_id)
        data = np.random.random_sample(shape)
        transformed = model.pipeline.transform(data)
        inverted = transformed.inverse_transform()
        return data, transformed.data, inverted

    def test_one_dimensional(self):
        data, transformed, inverted = self.do_setup((10,))
        self.assertEqual((10, 1), transformed.shape)
        self.assertEqual((10,), inverted.shape)
        np.testing.assert_allclose(data.ravel(), transformed.ravel())
        np.testing.assert_allclose(data.ravel(), inverted.ravel())

    def test_two_dimensional(self):
        data, transformed, inverted = self.do_setup((10, 10))
        self.assertEqual((10, 10), transformed.shape)
        self.assertEqual((10, 10), inverted.shape)
        np.testing.assert_allclose(data.ravel(), transformed.ravel())
        np.testing.assert_allclose(data.ravel(), inverted.ravel())

    def test_three_dimensional(self):
        data, transformed, inverted = self.do_setup((10, 10, 10))
        self.assertEqual((10, 100), transformed.shape)
        self.assertEqual((10, 10, 10), inverted.shape)
        np.testing.assert_allclose(data.ravel(), transformed.ravel())
        np.testing.assert_allclose(data.ravel(), inverted.ravel())



import numpy as np
from util import sigmoid, stochastic_binary as sb
import hashlib


class Rbm(object):
    """
    A binary->binary
    `Restricted Boltzmann Machine <http://www.cs.toronto.edu/~hinton/absps/guideTR.pdf>`_.

    Adapted from matlab code written by Geoff Hinton, available
    `here <http://www.cs.toronto.edu/~hinton/code/rbm.m>`_

    Justifications for initialization parameters can be found
    `here <http://www.cs.toronto.edu/~hinton/absps/guideTR.pdf>`_

    TODO: More in-depth explanation of RBMs
    """

    def __init__(
            self,
            indim,
            hdim,
            learning_rate=0.1,
            weight_decay=.002,
            initial_momentum=.5,
            final_momentum=.9,
            sparsity_target=.01,
            sparsity_decay=.9,
            sparsity_cost=.001):

        """__init__

        :param indim: The number of units in the visible layer, i.e., the \
        dimension of training data vectors

        :param hdim: The number of units in the hidden layer, i.e., the \
        dimension of the representation

        :param learning_rate:  The size of "step" to take when adjusting weights \
        during the update step

        :param weight_decay: A value which discourages large weights

        :param initial_momentum: The amount by which momentum from previous \
        updates decays from epoch to epoch, for the first five epochs.

        :param final_momentum: The amount by which momentum from previous \
        updates decays from epoch to epoch, after the first five epochs.

        :param sparsity_target:  The average number of hidden units that should \
        be "on" at once.

        :param sparsity_decay: The amount by which the sparsity penalty decays from \
        epoch to epoch

        :param sparsity_cost: The importance of the sparsity target.  Higher numbers \
        increase the importance of sparsity to the learning objective.
        """

        super(Rbm, self).__init__()

        # IO
        self._indim = indim
        self._hdim = hdim

        # sparsity
        self._sparsity_target = sparsity_target
        self._sparsity_decay = sparsity_decay
        self._sparsity_cost = sparsity_cost

        # initialize biases
        self._vbias = np.zeros(self._indim)
        self._hbias = np.zeros(self._hdim)
        if self._sparsity_target:
            self._hbias[:] = \
                np.log(self._sparsity_target / (1 - self._sparsity_target))

        # initialize the weights to a normal
        # distribution with 0 mean and 0.01 std
        self._weights = np.random.normal(0.0, .01, (self._indim, self._hdim))

        # learning rate, weight decay, momentum
        self._learning_rate = learning_rate
        self._weight_decay = weight_decay
        self._initial_momentum = initial_momentum
        self._final_momentum = final_momentum

    @property
    def version(self):
        h = hashlib.md5(self._weights)
        h.update(self._vbias)
        h.update(self._hbias)
        return h.hexdigest()

    @property
    def indim(self):
        """
        The size of the visible layer, i.e., the dimension of input data vectors
        """
        return self._indim

    @property
    def hdim(self):
        """
        The size of the hidden layer, i.e. the dimension of the learned
        representation
        """
        return self._hdim

    @property
    def dim(self):
        return self.hdim

    def _up(self, v):
        """
        Given a visible sample, return the pre-sigmoid and
        sigmoid activations of the hidden units
        """
        pre_sigmoid = np.dot(v, self._weights) + self._hbias
        return pre_sigmoid, sigmoid(pre_sigmoid)

    def _h_from_v(self, v):
        """
        Infer the state of the stochastic binary
        hidden units
        """
        ps, s = self._up(v)
        return ps, s, sb(s)

    def _down(self, h):
        """
        Given a hidden state, infer the activation of
        the visible units.
        """
        pre_sigmoid = np.dot(h, self._weights.T) + self._vbias
        sig = sigmoid(pre_sigmoid)
        return pre_sigmoid, sig, sb(sig)

    def _v_from_h(self, h):
        """
        At the moment, I'm following the pattern of the
        rbm in the theano tutorial, but this is really just
        a superfluous wrapper
        """
        ps, s, stoch = self._down(h)
        return ps, s, stoch

    def _gibbs_hvh(self, h):
        """
        One step of gibbs sampling, starting from the
        hidden layer
        """
        vps, vs, vstoch = self._v_from_h(h)
        ps, s, stoch = self._h_from_v(vs)
        return vs, ps, s, stoch

    def _positive_phase(self, inp):
        """
        positive phase sample (reality)
        get the pre-sigmoid, sigmoid, and stochastic
        binary activations of the hidden units, given
        the visible activations
        """
        ps, s, stoch = self._h_from_v(inp)
        posprod = np.dot(inp.T, s)
        pos_h_act = s.sum(axis=0)
        pos_v_act = inp.sum(axis=0)
        return stoch, posprod, pos_h_act, pos_v_act

    def _negative_phase(self, stoch):
        """
        negative phase sample (fantasy)
        """
        vs, gps, gs, gstoch = self._gibbs_hvh(stoch)
        negprod = np.dot(vs.T, gs)
        neg_h_act = gs.sum(axis=0)
        neg_v_act = vs.sum(axis=0)
        return vs, negprod, neg_h_act, neg_v_act

    def _update(self, inp, momentum, epoch, batch):
        """
        Compare the energy of the input with the energy
        of a fantasy produced by beginning with that input and
        doing one step of gibbs sampling

        Update the weights by multiplying the difference between
        the input and fantasy free energies by a learning rate
        """

        # positive phase (reality)
        stoch, posprod, pos_h_act, pos_v_act = self._positive_phase(inp)

        # negative phase (fantasy)
        v, negprod, neg_h_act, neg_v_act = self._negative_phase(stoch)

        # calculate reconstruction error
        error = np.sum(np.abs(inp - v) ** 2)

        # number of samples in this batch
        n = float(inp.shape[0])
        m = momentum
        lr = self._learning_rate
        wd = self._weight_decay

        # sparsity
        if self._sparsity_target is not None:
            current_sparsity = stoch.sum(0) / float(n)
            self._sparsity = (self._sparsity_decay * self._sparsity) + \
                             ((1 - self._sparsity_decay) * current_sparsity)
            sparse_penalty = self._sparsity_cost * \
                             (self._sparsity_target - self._sparsity)

        # update the weights
        self._wvelocity = (m * self._wvelocity) + \
                          lr * (
                              ((posprod - negprod) / n) - (wd * self._weights))
        self._weights += self._wvelocity
        if self._sparsity_target is not None:
            self._weights += sparse_penalty

        # update the visible biases
        self._vbvelocity = (m * self._vbvelocity) + \
                           ((lr / n) * (pos_v_act - neg_v_act))
        self._vbias += self._vbvelocity

        # update the hidden biases
        self._hbvelocity = (m * self._hbvelocity) + \
                           ((lr / n) * (pos_h_act - neg_h_act))
        self._hbias += self._hbvelocity
        if self._sparsity_target is not None:
            self._hbias += sparse_penalty

        return error

    def train(self, samples, stopping_condition):
        """train

        :param samples: A two-dimensional numpy array whose second dimension \
        should be equal to :py:meth:`Rbm.indim`

        :param stopping_condition: A callable which takes epoch and error as \
        arguments

        """

        batch_size = 100
        nbatches = int(len(samples) / batch_size)
        # If the number of samples isn't evenly divisible by batch size, we're
        # just going to shave off the end.  If the training set is sufficiently
        # large (hopefully in the tens or hundreds of thousands), this shouldn't
        # be a big deal.
        samples = samples[:nbatches * batch_size].reshape(
                (nbatches, batch_size, samples.shape[1]))
        # initialize some variables we'll
        # use during training, but then throw away
        self._wvelocity = np.zeros(self._weights.shape)
        self._vbvelocity = np.zeros(self._indim)
        self._hbvelocity = np.zeros(self._hdim)
        self._sparsity = np.zeros(self._hdim)

        epoch = 0
        error = 99999
        nbatches = len(samples)
        while not stopping_condition(epoch, error):
            if epoch < 5:
                mom = self._initial_momentum
            else:
                mom = self._final_momentum
            batch = 0
            while batch < nbatches and not stopping_condition(epoch, error):
                error = self._update(samples[batch], mom, epoch, batch)
                batch += 1
                print 'Epoch %i, Batch %i, Error %1.4f' % (epoch, batch, error)
            epoch += 1

        # get rid of the "temp" training variables
        del self._wvelocity
        del self._vbvelocity
        del self._hbvelocity
        del self._sparsity

    def activate(self, inp, binarize=True):
        """
        Activate the net using the visible sample.
        Threshold the probabilities that the hidden
        units are on.
        """
        hps, hs, hstoch = self._h_from_v(inp)
        vps, vs, vstoch = self._v_from_h(hs)
        if binarize:
            vs[vs > .5] = 1
            vs[vs <= .5] = 0
        return vs

    def fromfeatures(self, features, binarize=True):
        """
        Using the hidden or latent variables, return the reconstructed output
        """
        vps, vs, vstoch = self._v_from_h(features)
        if binarize:
            vs[vs > .5] = 1
            vs[vs <= .5] = 0
        return vs

    # TODO: Is this the correct implementation for both Rbm and LinearRbm?
    def __call__(self, data, binarize=True):
        """__call__

        :param data: A two-dimensional numpy array of input data vectors

        :returns: The activations of the hidden layer for each input example
        """

        ps, s, stoch = self._h_from_v(data)
        if binarize:
            s[s > .5] = 1
            s[s <= .5] = 0
        return s


class RealValuedRbm(Rbm):
    """
    A Restricted Boltzmann Machine with gaussian visible units, as described
    in section 13.2 "Gaussian visible units",
    `here <http://www.cs.toronto.edu/~hinton/absps/guideTR.pdf>`_.

    When learning a deep-belief network, an instance of this class usually \
    constitutes the first layer, as the raw data is typically real-valued, \
    instead of binary.
    """

    def __init__(self, indim, hdim, sparsity_target=.01, learning_rate=.001):
        """__init__

        :param indim: The number of units in the visible layer, i.e., the \
        dimension of training data vectors

        :param hdim: The number of units in the hidden layer, i.e., the \
        dimension of the representation

        :param sparsity_target:  The average number of hidden units that should \
        be "on" at once.

        :param learning_rate:  The size of "step" to take when adjusting weights \
        during the update step.  The default :code:`learning_rate` is lower than \
        in the :py:class:`Rbm` because *"The learning rate needs to be about one \
        or two orders of magnitude smaller than when using binary visible units \
        ...A smaller learning rate is required because there is no upper bound \
        to the size of a component in the reconstruction and if one component \
        becomes very large, the weights emanating from it will get a very big \
        learning signal".*, from section 13.2 "Gaussian visible units", available \
        `here <http://www.cs.toronto.edu/~hinton/absps/guideTR.pdf>`_

        """
        super(RealValuedRbm, self).__init__(
                indim,
                hdim,
                learning_rate=learning_rate,
                sparsity_target=sparsity_target)

    def _gibbs_hvh(self, h):
        """
        One step of gibbs sampling, starting from the
        hidden layer
        """
        vps, vs, vstoch = self._v_from_h(h)
        # use the actual value of the visible units,
        # instead of the value passed through a sigmoid function
        ps, s, stoch = self._h_from_v(vps)
        return vps, ps, s, stoch

    def activate(self, inp):
        hps, hs, hstoch = self._h_from_v(inp)
        vps, vs, vstoch = self._v_from_h(hs)
        return vps



import frequencyscale
from zounds.core import Dimension


class FrequencyDimension(Dimension):
    def __init__(self, scale):
        super(FrequencyDimension, self).__init__()
        self.scale = scale

    def modified_dimension(self, size, windowsize, stepsize=None):
        raise NotImplementedError()

    def metaslice(self, index, size):
        return FrequencyDimension(self.scale[index])

    def integer_based_slice(self, index):
        if not isinstance(index, frequencyscale.FrequencyBand):
            return index

        return self.scale.get_slice(index)

    def validate(self, size):
        msg = 'scale and array size must match, ' \
            'but were scale: {self.scale.n_bands},  array size: {size}'

        if size != len(self.scale):
            raise ValueError(msg.format(**locals()))

    def __eq__(self, other):
        return self.scale == other.scale

    def __str__(self):
        return 'FrequencyDimension(scale={self.scale})'.format(**locals())

    def __repr__(self):
        return self.__str__()


import unittest2
from frequencydimension import \
    FrequencyDimensionEncoder, FrequencyDimensionDecoder
from zounds.spectral import \
    FrequencyDimension, FrequencyScale, FrequencyBand, LinearScale


class FrequencyDimensionTests(unittest2.TestCase):
    def setUp(self):
        self.encoder = FrequencyDimensionEncoder()
        self.decoder = FrequencyDimensionDecoder()

    def test_can_round_trip(self):
        band = FrequencyBand(20, 20000)
        scale = FrequencyScale(band, 50)
        dim = FrequencyDimension(scale)
        encoded = self.encoder.encode(dim)
        decoded = self.decoder.decode(encoded)
        self.assertIsInstance(decoded, FrequencyDimension)
        self.assertEqual(scale, decoded.scale)

    def test_can_round_trip_specific_scale_type(self):
        band = FrequencyBand(20, 20000)
        scale = LinearScale(band, 50)
        dim = FrequencyDimension(scale)
        encoded = self.encoder.encode(dim)
        decoded = self.decoder.decode(encoded)
        self.assertIsInstance(decoded.scale, LinearScale)



import unittest2
from dimension import DimensionEncoder, DimensionDecoder
from zounds.timeseries import TimeDimension, Seconds, Milliseconds
from zounds.spectral import FrequencyBand, FrequencyScale, FrequencyDimension
from zounds.core import IdentityDimension


class RoundTripTests(unittest2.TestCase):

    def roundtrip(self, o):
        encoder = DimensionEncoder()
        decoder = DimensionDecoder()
        encoded = list(encoder.encode(o))
        return list(decoder.decode(encoded))

    def test_can_round_trip_single_identity_dimension(self):
        original = [IdentityDimension()]
        restored = self.roundtrip(original)
        self.assertSequenceEqual(original, restored)

    def test_can_round_trip_mixed_dimensions(self):
        original = [
            IdentityDimension(),
            TimeDimension(Seconds(1), Milliseconds(500)),
            FrequencyDimension(FrequencyScale(FrequencyBand(100, 1000), 10))
        ]
        restored = self.roundtrip(original)
        self.assertSequenceEqual(original, restored)



from scipy.cluster.vq import kmeans
from featureflow import Node
from preprocess import PreprocessResult, Preprocessor
from rbm import Rbm, RealValuedRbm


class KMeans(Preprocessor):
    def __init__(self, centroids=None, needs=None):
        super(KMeans, self).__init__(needs=needs)
        self._centroids = centroids

    def _forward_func(self):
        def x(d, codebook=None):
            import numpy as np
            from scipy.spatial.distance import cdist
            from zounds.core import ArrayWithUnits, IdentityDimension
            l = d.shape[0]
            dist = cdist(d, codebook)
            best = np.argmin(dist, axis=1)
            feature = np.zeros((l, len(codebook)), dtype=np.uint8)
            feature[np.arange(l), best] = 1
            try:
                return ArrayWithUnits(
                        feature, [d.dimensions[0], IdentityDimension()])
            except AttributeError:
                return feature

        return x

    def _backward_func(self):
        def x(d, codebook=None):
            import numpy as np
            indices = np.where(d == 1)
            return codebook[indices[1]]

        return x

    def _process(self, data):
        data = self._extract_data(data)
        codebook, _ = kmeans(data, self._centroids)
        op = self.transform(codebook=codebook)
        inv_data = self.inversion_data(codebook=codebook)
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data, op, inversion_data=inv_data, inverse=inv, name='KMeans')


class BaseRbm(Preprocessor):
    def __init__(
            self,
            cls=None,
            hdim=None,
            sparsity_target=0.01,
            epochs=100,
            learning_rate=0.1,
            needs=None):
        super(BaseRbm, self).__init__(needs=needs)
        self._hdim = hdim
        self._sparsity = sparsity_target
        self._learning_rate = learning_rate
        self._epochs = epochs
        self._cls = cls

    def _forward_func(self):
        def x(d, rbm=None):
            return rbm(d)

        return x

    def _backward_func(self):
        def x(d, rbm=None):
            return rbm.fromfeatures(d, binarize=False)

        return x

    def _process(self, data):
        data = self._extract_data(data)
        rbm = self._cls(
                indim=data.shape[1],
                hdim=self._hdim,
                sparsity_target=self._sparsity,
                learning_rate=self._learning_rate)
        rbm.train(data, lambda epoch, error: epoch > self._epochs)
        op = self.transform(rbm=rbm)
        inv_data = self.inversion_data(rbm=rbm)
        inv = self.inverse_transform()
        data = op(data)
        yield PreprocessResult(
                data,
                op,
                inversion_data=inv_data,
                inverse=inv,
                name=self.__class__.__name__)


class BinaryRbm(BaseRbm):
    def __init__(
            self, sparsity_target=0.01,
            hdim=None,
            epochs=100,
            learning_rate=0.1,
            needs=None):
        super(BinaryRbm, self).__init__(
                cls=Rbm,
                hdim=hdim,
                sparsity_target=sparsity_target,
                epochs=epochs,
                learning_rate=learning_rate,
                needs=needs)


class LinearRbm(BaseRbm):
    def __init__(
            self,
            hdim=None,
            sparsity_target=0.01,
            epochs=100,
            learning_rate=0.001,
            needs=None):
        super(LinearRbm, self).__init__(
                cls=RealValuedRbm,
                hdim=hdim,
                sparsity_target=sparsity_target,
                epochs=epochs,
                learning_rate=learning_rate,
                needs=needs)


class Learned(Node):
    def __init__(self, learned=None, needs=None):
        super(Learned, self).__init__(needs=needs)
        self._learned = learned

    @property
    def version(self):
        return self._learned.pipeline.version

    def _process(self, data):
        transformed = self._learned.pipeline.transform(data).data
        yield transformed



from __future__ import division
import numpy as np
import bisect


# TODO: What commonalities can be factored out of this class and TimeSlice?
# TODO: Deprecate stuff in psychoacoustics.py in favor of these classes
class FrequencyBand(object):
    """
    Represents an interval, or band of frequencies in hertz (cycles per second)
    """

    def __init__(self, start_hz, stop_hz):
        """
        :param start_hz: The lower bound for the frequency band, in hertz
        :param stop_hz: The upper bound for the frequency band, in hertz
        :return: a new FrequencyBand instance
        """
        super(FrequencyBand, self).__init__()
        if stop_hz <= start_hz:
            raise ValueError('stop_hz must be greater than start_hz')
        self.stop_hz = stop_hz
        self.start_hz = start_hz

    def __eq__(self, other):
        try:
            return \
                self.start_hz == other.start_hz \
                and self.stop_hz == other.stop_hz
        except AttributeError:
            return super(FrequencyBand, self).__eq__(other)

    @staticmethod
    def from_center(center_hz, bandwidth_hz):
        half_bandwidth = bandwidth_hz / 2
        return FrequencyBand(
                center_hz - half_bandwidth, center_hz + half_bandwidth)

    @property
    def bandwidth(self):
        return self.stop_hz - self.start_hz

    @property
    def center_frequency(self):
        return self.start_hz + (self.bandwidth / 2)

    def __repr__(self):
        return '''FrequencyBand(
start_hz={start_hz},
stop_hz={stop_hz},
center={center},
bandwidth={bandwidth})'''.format(
                start_hz=self.start_hz,
                stop_hz=self.stop_hz,
                center=self.center_frequency,
                bandwidth=self.bandwidth)


class FrequencyScale(object):
    """
    Represents a set of frequency bands with monotonically increasing start
    frequencies
    """

    def __init__(self, frequency_band, n_bands, always_even=False):
        """
        :param frequency_band: A wide band that defines the boundaries for the
        entire scale.  E.g., one might want to generate a scale spanning the
        full range of (normal) human hearing by starting with
        FrequencyBand(20, 20000)
        :param n_bands: The total number of individual bands in the scale
        :return: a new FrequencyScale instance
        """
        super(FrequencyScale, self).__init__()
        self.always_even = always_even
        self.n_bands = n_bands
        self.frequency_band = frequency_band
        self._bands = None

    @property
    def bands(self):
        if self._bands is None:
            self._bands = self._compute_bands()
        return self._bands

    def _compute_bands(self):
        raise NotImplementedError()

    def __len__(self):
        return self.n_bands

    @property
    def center_frequencies(self):
        return (band.center_frequency for band in self)

    @property
    def bandwidths(self):
        return (band.bandwidth for band in self)

    @property
    def start_hz(self):
        return self.frequency_band.start_hz

    @property
    def stop_hz(self):
        return self.frequency_band.stop_hz

    def get_slice(self, frequency_band):
        """
        Given a frequency band, and a frequency dimension comprised of n_samples,
        return a slice using integer indices that may be used to extract only
        the frequency samples that intersect with the frequency band
        :param frequency_band: The range of frequencies for which a slice should
        be produced
        :return: a slice
        """
        starts = [b.start_hz for b in self.bands]
        stops = [b.stop_hz for b in self.bands]
        start_index = bisect.bisect_left(stops, frequency_band.start_hz)
        stop_index = bisect.bisect_left(starts, frequency_band.stop_hz)

        if self.always_even and (stop_index - start_index) % 2:
            # KLUDGE: This is simple, but it may make sense to choose move the
            # upper *or* lower bound, based on which one introduces a lower
            # error
            stop_index += 1
        return slice(start_index, stop_index)

    def __eq__(self, other):
        return \
            self.__class__ == other.__class__ \
            and self.frequency_band == other.frequency_band \
            and self.n_bands == other.n_bands

    def __iter__(self):
        return iter(self.bands)

    def __getitem__(self, index):
        try:
            # index is an integer or slice
            bands = self.bands[index]
        except TypeError:
            # index is a frequency band
            bands = self.bands[self.get_slice(index)]

        try:
            freq_band = FrequencyBand(bands[0].start_hz, bands[-1].stop_hz)
            return self.__class__(freq_band, len(bands))
        except TypeError:
            # we've already got an individual band
            return bands

    def __str__(self):
        cls = self.__class__.__name__
        return '{cls}(band={self.frequency_band}, n_bands={self.n_bands})'\
            .format(**locals())

    def __repr__(self):
        return self.__str__()


class LinearScale(FrequencyScale):
    """
    A linear frequency scale with constant bandwidth.  Appropriate for use
    with transforms whose coefficients also lie on a linear frequency scale,
    e.g. the FFT or DCT transforms.
    """

    def __init__(self, frequency_band, n_bands, always_even=False):
        super(LinearScale, self).__init__(frequency_band, n_bands, always_even)

    @staticmethod
    def from_sample_rate(sample_rate, n_bands, always_even=False):
        fb = FrequencyBand(0, sample_rate.nyquist)
        return LinearScale(fb, n_bands, always_even=always_even)

    def _compute_bands(self):
        freqs = np.linspace(self.start_hz, self.stop_hz, self.n_bands)
        # constant, non-overlapping bandwidth
        bandwidth = freqs[1] - freqs[0]
        return tuple(FrequencyBand(f, f + bandwidth) for f in freqs)


class LogScale(FrequencyScale):
    def __init__(self, frequency_band, n_bands, always_even=False):
        super(LogScale, self).__init__(
                frequency_band, n_bands, always_even=always_even)

    def _compute_bands(self):
        center_freqs = np.logspace(
                np.log10(self.start_hz),
                np.log10(self.stop_hz),
                self.n_bands + 1)
        # variable bandwidth
        bandwidths = np.diff(center_freqs)
        return tuple(FrequencyBand.from_center(cf, bw)
                for (cf, bw) in zip(center_freqs[:-1], bandwidths))

class BarkScale(FrequencyScale):
    def __init__(self, frequency_band, n_bands):
        super(BarkScale, self).__init__(frequency_band, n_bands)


class MelScale(FrequencyScale):
    def __init__(self, frequency_band, n_bands):
        super(MelScale, self).__init__(frequency_band, n_bands)



import unittest2
from dimensions import IdentityDimension, Dimension


class IdentityDimensionTests(unittest2.TestCase):
    def test_equal(self):
        d1 = IdentityDimension()
        d2 = IdentityDimension()
        self.assertEqual(d1, d2)

    def test_not_equal(self):
        d1 = Dimension()
        d2 = IdentityDimension()
        self.assertNotEqual(d1, d2)



import featureflow as ff


def simple_lmdb_settings(path, map_size=1e9):
    """
    Creates a decorator that can be used to configure sane default LMDB
    persistence settings for a model

    :param path: The path where the LMDB database files will be created
    :param map_size: The amount of space to allot for the LMDB database
    :return: a decorator that configures persistence settings
    """

    def decorator(cls):
        class Settings(ff.PersistenceSettings):
            id_provider = ff.UuidProvider()
            key_builder = ff.StringDelimitedKeyBuilder()
            database = ff.LmdbDatabase(
                    path, key_builder=key_builder, map_size=map_size)

        class Model(cls, Settings):
            pass

        return Model

    return decorator


def simple_object_storage_settings(container, region, username, api_key):
    def decorator(cls):
        class Settings(ff.PersistenceSettings):
            id_provider = ff.UuidProvider()
            key_builder = ff.StringDelimitedKeyBuilder()
            database = ff.ObjectStoreDatabase(
                container, username, api_key, region, key_builder=key_builder)

        class Model(cls, Settings):
            pass

        return Model

    return decorator


def simple_in_memory_settings(cls):
    """
    Decorator that returns a class that "persists" data in-memory.  Mostly
     useful for testing
    :param cls: the class whose features should be persisted in-memory
    :return: A new class that will persist features in memory
    """

    class Settings(ff.PersistenceSettings):
        id_provider = ff.UuidProvider()
        key_builder = ff.StringDelimitedKeyBuilder()
        database = ff.InMemoryDatabase(key_builder=key_builder)

    class Model(cls, Settings):
        pass

    return Model



from __future__ import division
from duration import Picoseconds
from collections import namedtuple
import numpy as np
from duration import Seconds

Stride = namedtuple('Stride', ['frequency', 'duration'])


class SampleRate(object):
    def __init__(self, frequency, duration):
        self.frequency = frequency
        self.duration = duration
        super(SampleRate, self).__init__()

    def __str__(self):
        f = self.frequency / Seconds(1)
        d = self.duration / Seconds(1)
        return '{self.__class__.__name__}(f={f}, d={d})'.format(**locals())

    def __repr__(self):
        return self.__str__()

    def __iter__(self):
        return iter((self.frequency, self.duration))

    def __len__(self):
        return 2

    def __eq__(self, other):
        return \
            self.frequency == other.frequency \
            and self.duration == other.duration

    @property
    def overlap(self):
        return self.duration - self.frequency

    def __mul__(self, other):
        try:
            if len(other) == 1:
                other *= 2
        except TypeError:
            other = (other, other)

        freq = self.frequency * other[0]
        duration = (self.frequency * other[1]) + self.overlap
        new = SampleRate(freq, duration)
        print self, new
        return new

    def discrete_samples(self, ts):
        """
        Compute the frequency and duration in discrete samples
        :param ts: A ConstantRateTimeSeries instance
        :return: A tuple, representing the frequency and duration in discrete
        samples, respectively
        """
        td = ts.dimensions[0]
        windowsize = np.round((self.duration - td.overlap) / td.frequency)
        stepsize = np.round(self.frequency / td.frequency)
        return int(stepsize), int(windowsize)


class AudioSampleRate(SampleRate):
    def __init__(self, samples_per_second, suggested_window, suggested_hop):
        self.suggested_hop = suggested_hop
        self.suggested_window = suggested_window
        self.one_sample = Picoseconds(int(1e12)) // samples_per_second
        super(AudioSampleRate, self).__init__(self.one_sample, self.one_sample)

    def __int__(self):
        return self.samples_per_second

    @property
    def samples_per_second(self):
        return int(Picoseconds(int(1e12)) / self.frequency)

    @property
    def nyquist(self):
        return self.samples_per_second // 2

    def half_lapped(self):
        return SampleRate(
                self.one_sample * self.suggested_hop,
                self.one_sample * self.suggested_window)


class SR96000(AudioSampleRate):
    def __init__(self):
        super(SR96000, self).__init__(96000, 4096, 2048)


class SR48000(AudioSampleRate):
    def __init__(self):
        super(SR48000, self).__init__(48000, 2048, 1024)


class SR44100(AudioSampleRate):
    def __init__(self):
        super(SR44100, self).__init__(44100, 2048, 1024)


class SR22050(AudioSampleRate):
    def __init__(self):
        super(SR22050, self).__init__(22050, 1024, 512)


class SR11025(AudioSampleRate):
    def __init__(self):
        super(SR11025, self).__init__(11025, 512, 256)


_samplerates = (SR96000(), SR48000(), SR44100(), SR22050(), SR11025())


def audio_sample_rate(samples_per_second):
    for sr in _samplerates:
        if samples_per_second == sr.samples_per_second:
            return sr
    raise ValueError(
            '{samples_per_second} is an invalid sample rate'.format(**locals()))


class HalfLapped(SampleRate):
    def __init__(self, window_at_44100=2048, hop_at_44100=1024):
        one_sample_at_44100 = Picoseconds(int(1e12)) / 44100.
        window = one_sample_at_44100 * window_at_44100
        step = one_sample_at_44100 * hop_at_44100
        super(HalfLapped, self).__init__(step, window)



import numpy as np
import unittest2
from zounds.basic import resampled
from zounds.util import simple_in_memory_settings
from zounds.timeseries import \
    SR11025, SR22050, SR44100, Seconds, Milliseconds, Picoseconds, \
    AudioSamples, TimeSlice, TimeDimension
from zounds.timeseries.samplerate import SampleRate, HalfLapped
from zounds.synthesize import \
    SineSynthesizer, DCTIVSynthesizer, MDCTSynthesizer, NoiseSynthesizer
from zounds.spectral import \
    SlidingWindow, DCTIV, MDCT, FFT, SpectralCentroid, OggVorbisWindowingFunc, \
    SpectralFlatness
from zounds.persistence import ArrayWithUnitsFeature
from zounds.core import ArrayWithUnits
import scipy


class SpectralFlatnessTests(unittest2.TestCase):
    def setUp(self):
        self.samplerate = SR22050()
        rs = resampled(resample_to=self.samplerate)

        wscheme = HalfLapped()

        @simple_in_memory_settings
        class Document(rs):
            windowed = ArrayWithUnitsFeature(
                    SlidingWindow,
                    wscheme=wscheme,
                    wfunc=OggVorbisWindowingFunc(),
                    needs=rs.resampled,
                    store=False)

            fft = ArrayWithUnitsFeature(
                    FFT,
                    needs=windowed,
                    store=False)

            flatness = ArrayWithUnitsFeature(
                    SpectralFlatness,
                    needs=fft,
                    store=True)

        # create a pure sine wave that fades out
        ss = SineSynthesizer(self.samplerate)
        sine = ss.synthesize(Seconds(5), [440.])
        sine_envelope = np.linspace(1.0, 0.0, len(sine))
        sine *= sine_envelope

        # create noise
        ns = NoiseSynthesizer(self.samplerate)
        noise = ns.synthesize(Seconds(5))
        noise_envelope = np.linspace(0.0, 1.0, len(noise))
        noise *= noise_envelope

        # mix the sine wave and noise together
        self.audio = sine + noise

        _id = Document.process(meta=self.audio.encode())
        self.doc = Document(_id)

    def test_has_correct_type(self):
        self.assertIsInstance(self.doc.flatness, ArrayWithUnits)

    def test_has_correct_duration(self):
        self.assertAlmostEqual(
                self.audio.dimensions[0].end_seconds,
                self.doc.flatness.dimensions[0].end_seconds,
                delta=0.02)

    def test_has_correct_dimensions(self):
        self.assertEqual(1, len(self.doc.flatness.dimensions))
        self.assertIsInstance(self.doc.flatness.dimensions[0], TimeDimension)

    def test_flatness_is_monotonically_increasing(self):
        chunked = self.doc.flatness \
            .sliding_window((TimeSlice(Seconds(1)),)) \
            .mean(axis=1)
        diff = np.diff(chunked)
        self.assertTrue(np.all(diff >= 0))


class SpectralCentroidTests(unittest2.TestCase):
    def setUp(self):
        self.samplerate = SR44100()
        rs = resampled(resample_to=self.samplerate)

        wscheme = HalfLapped()

        @simple_in_memory_settings
        class Document(rs):
            windowed = ArrayWithUnitsFeature(
                    SlidingWindow,
                    wscheme=wscheme,
                    wfunc=OggVorbisWindowingFunc(),
                    needs=rs.resampled,
                    store=False)

            fft = ArrayWithUnitsFeature(
                    FFT,
                    needs=windowed,
                    store=False)

            centroid = ArrayWithUnitsFeature(
                    SpectralCentroid,
                    needs=fft,
                    store=True)

        ss = SineSynthesizer(self.samplerate)
        chunks = \
            [ss.synthesize(Seconds(1), [440 * i]) for i in xrange(1, 6)]
        self.audio = \
            AudioSamples(ArrayWithUnits.concat(chunks), self.samplerate)

        _id = Document.process(meta=self.audio.encode())
        self.doc = Document(_id)

    def test_has_correct_type(self):
        self.assertIsInstance(self.doc.centroid, ArrayWithUnits)

    def test_has_correct_dimensions(self):
        self.assertEqual(1, len(self.doc.centroid.dimensions))

    def test_has_correct_duration(self):
        self.assertAlmostEqual(
                self.audio.dimensions[0].end_seconds,
                self.doc.centroid.dimensions[0].end_seconds,
                delta=0.02)

    def test_centroid_is_monotonically_increasing(self):
        chunked = self.doc.centroid \
            .sliding_window((TimeSlice(Seconds(1)),)) \
            .mean(axis=1)
        diff = np.diff(chunked)
        self.assertTrue(np.all(diff >= 0))


class MDCTTests(unittest2.TestCase):
    def setUp(self):
        self.samplerate = SR11025()
        rs = resampled(resample_to=self.samplerate)

        wscheme = HalfLapped()

        @simple_in_memory_settings
        class Document(rs):
            windowed = ArrayWithUnitsFeature(
                    SlidingWindow,
                    wscheme=wscheme,
                    needs=rs.resampled,
                    store=False)

            mdct = ArrayWithUnitsFeature(
                    MDCT,
                    needs=windowed,
                    store=True)

        ss = SineSynthesizer(self.samplerate)
        self.audio = ss.synthesize(Seconds(5), [440., 660., 880.])

        _id = Document.process(meta=self.audio.encode())
        self.doc = Document(_id)

    def test_has_correct_duration(self):
        self.assertAlmostEqual(
                self.audio.dimensions[0].end_seconds,
                self.doc.mdct.dimensions[0].end_seconds,
                delta=0.02)

    def test_perfect_reconstruction_using_overlap_add(self):
        synth = SineSynthesizer(SR22050())
        audio = synth.synthesize(Seconds(10), [440., 660., 880.])
        windowed = audio.sliding_window(
                (TimeSlice(Seconds(1)),),
                (TimeSlice(Milliseconds(500)),))

        mdct = MDCT()

        coeffs = list(mdct._process(windowed * OggVorbisWindowingFunc()))[0]

        mdct_synth = MDCTSynthesizer()
        recon = mdct_synth.synthesize(coeffs)

        # take a slice, so we can ignore boundary conditions
        slce = TimeSlice(start=Seconds(1), duration=Seconds(8))

        np.testing.assert_almost_equal(recon[slce], audio[slce], decimal=4)

    def test_is_correct_type(self):
        self.assertIsInstance(self.doc.mdct, ArrayWithUnits)

    def test_has_correct_nyquist_frequency(self):
        freq_dim = self.doc.mdct.dimensions[1]
        self.assertEqual(self.samplerate.nyquist, freq_dim.scale.stop_hz)

    def test_reconstruction(self):
        ds = MDCTSynthesizer()
        recon = ds.synthesize(self.doc.mdct)

        orig_seconds = self.audio.dimensions[0].end_seconds
        recon_seconds = recon.dimensions[0].end_seconds
        self.assertAlmostEqual(orig_seconds, recon_seconds, delta=0.02)

        # ensure that both the original and reconstruction have the exact
        # same length in samples, so that we can easily compare spectral peaks
        l = min(len(self.audio), len(recon))

        self.assertEqual(self.audio.samplerate, recon.samplerate)
        orig_fft = abs(np.fft.rfft(self.audio[:l]))
        recon_fft = abs(np.fft.rfft(recon[:l]))
        orig_peaks = set(np.argsort(orig_fft)[-3:])
        recon_peaks = set(np.argsort(recon_fft)[-3:])
        # ensure that the original and reconstruction have the same three
        # spectral peaks
        self.assertEqual(orig_peaks, recon_peaks)


class DCTIVTests(unittest2.TestCase):
    def setUp(self):
        self.samplerate = SR22050()
        rs = resampled(resample_to=self.samplerate)

        window_size = Picoseconds(int(1e12))
        wscheme = SampleRate(window_size, window_size)

        @simple_in_memory_settings
        class Document(rs):
            windowed = ArrayWithUnitsFeature(
                    SlidingWindow,
                    wscheme=wscheme,
                    needs=rs.resampled,
                    store=False)

            dct = ArrayWithUnitsFeature(
                    DCTIV,
                    needs=windowed,
                    store=True)

        ss = SineSynthesizer(self.samplerate)
        self.audio = ss.synthesize(Seconds(5), [440., 660., 880.])

        _id = Document.process(meta=self.audio.encode())
        self.doc = Document(_id)

    @unittest2.skip
    def test_perfect_reconstruction(self):
        synth = SineSynthesizer(SR22050())
        audio = synth.synthesize(Seconds(1), [440., 660., 880.])
        node = DCTIV()
        coeffs = node._process_raw(audio[None, :])
        recon = node._process_raw(coeffs)[0]

        # see? it's do-able w/ dct
        dct_coeffs = scipy.fftpack.dct(audio[None, :], norm='ortho')
        inverse_dct = scipy.fftpack.idct(dct_coeffs, norm='ortho')[0]

        np.testing.assert_almost_equal(inverse_dct, audio, decimal=4)
        np.testing.assert_almost_equal(recon, audio, decimal=4)

    def test_is_correct_type(self):
        self.assertIsInstance(self.doc.dct, ArrayWithUnits)

    def test_has_correct_nyquist_frequency(self):
        freq_dim = self.doc.dct.dimensions[-1]
        self.assertEqual(self.samplerate.nyquist, freq_dim.scale.stop_hz)

    def test_reconstruction(self):
        ds = DCTIVSynthesizer()
        recon = ds.synthesize(self.doc.dct)

        orig_seconds = self.audio.dimensions[0].end_seconds
        recon_seconds = recon.dimensions[0].end_seconds
        self.assertAlmostEqual(orig_seconds, recon_seconds, delta=0.02)

        # ensure that both the original and reconstruction have the exact
        # same length in samples, so that we can easily compare spectral peaks
        l = min(len(self.audio), len(recon))

        self.assertEqual(self.audio.samplerate, recon.samplerate)
        orig_fft = abs(np.fft.rfft(self.audio[:l]))
        recon_fft = abs(np.fft.rfft(recon[:l]))
        orig_peaks = set(np.argsort(orig_fft)[-3:])
        recon_peaks = set(np.argsort(recon_fft)[-3:])
        # ensure that the original and reconstruction have the same three
        # spectral peaks
        self.assertEqual(orig_peaks, recon_peaks)



import tornado
from baseapp import BaseZoundsApp, RequestContext
import base64
import urllib


class ZoundsSearch(BaseZoundsApp):
    def __init__(
            self,
            base_path=r'/zounds/',
            model=None,
            visualization_feature=None,
            audio_feature=None,
            search=None,
            n_results=10):
        super(ZoundsSearch, self).__init__(
                base_path=base_path,
                model=model,
                visualization_feature=visualization_feature,
                audio_feature=audio_feature,
                html='search.html')
        self.n_results = n_results
        self.search = search

    def custom_routes(self):
        return [
            (r'/zounds/search', self.search_handler())
        ]

    def search_handler(self):
        app = self

        class SearchHandler(tornado.web.RequestHandler):
            def get(self):
                b64_encoded_query = urllib.unquote(
                        self.get_argument('query', default=''))
                if b64_encoded_query:
                    binary_query = base64.b64decode(b64_encoded_query)
                    query = app.search.decode_query(binary_query)
                    results = app.search.search(query, n_results=app.n_results)
                else:
                    results = app.search.random_search(n_results=app.n_results)
                context = RequestContext(value=results)
                output = app.serialize(context)
                self.set_header('Content-Type', output.content_type)
                self.write(output.data)

        return SearchHandler



import unittest2
from handy import tuplify


class TuplifyTests(unittest2.TestCase):
    def test_tuple_from_tuple(self):
        self.assertEqual((1, 2, 3), tuplify((1, 2, 3)))

    def test_tuple_from_list(self):
        self.assertEqual((1, 2, 3), tuplify([1, 2, 3]))

    def test_tuple_from_integer(self):
        self.assertEqual((1,), tuplify(1))
